#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

from nomad.metainfo import (
    MSection, Package, Quantity, SubSection, Datetime)
from nomad.metainfo.metainfo import Reference, SectionProxy


m_package = Package(name='measurements')


class Sample(MSection):
    sample_id = Quantity(
        type=str, description='Identification number or signatures of the sample used.')

    name = Quantity(
        type=str, description='A human readable free text name for the sample.')

    description = Quantity(
        type=str, description='A description of the sample.')

    sample_state = Quantity(
        type=str, description='The physical state of the sample.')

    sample_temperature = Quantity(
        type=np.dtype(np.float64), unit='kelvin',
        description='The temperature of the sample during the measurement.')

    sample_microstructure = Quantity(
        type=str, description='The sample microstructure.')

    sample_constituents = Quantity(
        type=str, description='The constituents.')

    elements = Quantity(type=str, shape=["*"])
    chemical_formula = Quantity(type=str)


class Experiment(MSection):
    name = Quantity(
        type=str, description='A human readable free text name for the experiment.')

    description = Quantity(
        type=str, description='A description of the experiment.')

    steps = Quantity(
        type=str, shape=['*'], description='Human readable experiment steps.')

    sample = SubSection(section_def=Sample, description='The used sample.')
    sample_ref = Quantity(type=Reference(Sample.m_def), description='Reference to the used sample.')

    measurement = SubSection(
        section_def=SectionProxy('Measurement'),
        repeats=True,
        description='Measurements performed in this experiment.')


class Instrument(MSection):
    instrument_id = Quantity(
        type=str, description='Identification number or signatures of the instrument used.')

    name = Quantity(
        type=str, description='A human readable free text name for the instrument.')

    description = Quantity(
        type=str, description='A description of the instrument.')


class Measurement(MSection):
    measurement_id = Quantity(type=str)
    name = Quantity(type=str)
    description = Quantity(type=str)

    method_name = Quantity(type=str)
    method_abbreviation = Quantity(type=str)

    start_time = Quantity(
        type=Datetime, description='The datetime of the beginning of the measurement.')
    end_time = Quantity(
        type=Datetime, description='The datetime of the measurement end.')

    facility = Quantity(
        type=str, description='''
        Description of the facility (e.g. in full or an acronym) where
        the measurement was conducted.''')

    sample = SubSection(section_def=Sample, repeats=True)
    instrument = SubSection(section_def=Instrument, repeats=True)


class Spectrum(MSection):
    '''
    Generic spectrum data with energies and counts. May include additional
    channels.
    '''

    class SpectrumChannel(MSection):
        '''
        Provides the metadata for a generic additional spectrum channel. Do not
        use it for energy or count; they have their predefined channels.
        '''
        channel_id = Quantity(type=str)
        label = Quantity(type=str)
        unit = Quantity(type=str)

    def derive_n_values(self):
        if self.count is not None:
            return len(self.count)
        if self.energy is not None:
            return len(self.energy)
        else:
            return 0

    n_values = Quantity(type=int, derived=derive_n_values)
    count = Quantity(
        type=np.dtype(np.float64), shape=['n_values'],
        description='The count at each energy value, dimensionless')
    energy = Quantity(
        type=np.dtype(np.float64), shape=['n_values'], unit='J',
        description='The energy range of the spectrum')

    additional_channel_data = Quantity(
        type=np.dtype(np.float64), shape=['n_channels', 'n_values'],
        description='''
            Data from additional channels. The channels are described in `additional channels`.
        ''')
    additional_channels = SubSection(
        sub_section=SpectrumChannel, repeats=True,
        description='''
            Metadata for additional channels. The order is the same as the channel data
            appears in `additional_channel_data`.
        ''')
    n_additional_channels = Quantity(type=int, derived=lambda spectrum: len(spectrum.additional_channels))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
from datetime import datetime
from typing import List, Dict, Set, Any, Optional
from pydantic import BaseModel

from pymongo import ReplaceOne
from pymongo.database import Database, Collection
from pymongo.cursor import Cursor
from nomad import utils
from nomad.processing import ProcessStatus, Upload, Entry
from nomad.datamodel import Dataset
from nomad.parsing.parsers import parser_dict


_upload_keys_to_remove_v0 = (
    'published', 'upload_path', 'upload_time', 'temporary', 'joined')
_metadata_keys_to_flatten_v0 = (
    'pid', 'external_id', 'nomad_version', 'nomad_commit', 'comment',
    'references', 'datasets')
_metadata_keys_to_remove_v0 = (
    'upload_name', 'upload_time', 'uploader', 'published', 'license', 'with_embargo',
    'external_db', 'shared_with', 'coauthors', 'processed', 'domain', 'raw_id')


class _CollectionStatistics(BaseModel):
    collection_name: str
    total: int = 0
    converted: int = 0
    migrated: int = 0


class _UpgradeStatistics(BaseModel):
    uploads = _CollectionStatistics(collection_name='Uploads')
    entries = _CollectionStatistics(collection_name='Entries')
    datasets = _CollectionStatistics(collection_name='Datasets')
    dois = _CollectionStatistics(collection_name='DOIs')


class _DatasetCacheItem(BaseModel):
    converted_dataset_dict: Optional[Dict[str, Any]] = None
    converted_doi_dict: Optional[Dict[str, Any]] = None
    ready_to_commit: bool = False


def create_collections_if_needed(db_dst: Database):
    '''
    If the collections haven't yet been created, create them by calling .objects() on the
    MongoDocument class.
    '''
    if 'upload' not in db_dst.collection_names():
        Upload.objects()
    if 'entry' not in db_dst.collection_names():
        Entry.objects()
    if 'dataset' not in db_dst.collection_names():
        Dataset.m_def.a_mongo.objects()


def migrate_mongo_uploads(
        db_src: Database, db_dst: Database, uploads: Cursor, failed_ids_to_file: bool,
        upload_update: Dict[str, Any], entry_update: Dict[str, Any], overwrite: str,
        fix_problems: bool, dry: bool):
    ''' Converts and/or migrates an upload and all related entries and datasets. '''
    logger = utils.get_logger(__name__)

    number_of_uploads = uploads.count()
    print(f'Found {number_of_uploads} uploads to import.')
    src_entry_collection = db_src.calc if 'calc' in db_src.collection_names() else db_src.entry
    dataset_cache: Dict[str, _DatasetCacheItem] = {}
    stats = _UpgradeStatistics()
    stats.uploads.total = number_of_uploads
    count_treated = count_failures = count_processing = 0
    failed_and_skipped = []
    count_ignored = 0
    start_time = time.time()
    next_report_time = start_time + 60
    for upload_dict in uploads:
        upload_id = upload_dict.get('_id')
        try:
            if _is_processing(upload_dict):
                logger.warn(f'Skipping upload because it is currently processing', upload_id=upload_id)

                count_processing += 1
                failed_and_skipped.append(upload_id)
            else:
                upload_dict_dst = db_dst.upload.find_one({'_id': upload_id})
                if upload_dict_dst:
                    if overwrite == 'always':
                        assert not _is_processing(upload_dict_dst), 'Destination upload is processing'
                    elif overwrite == 'if-newer':
                        complete_time = upload_dict.get('complete_time', datetime.min)
                        complete_time_dst = upload_dict_dst.get('complete_time', datetime.min)
                        if _is_processing(upload_dict_dst) or complete_time <= complete_time_dst:
                            count_ignored += 1
                            continue
                    elif overwrite == 'never':
                        count_ignored += 1
                        continue

                entry_dicts, dataset_dicts, doi_dicts = _convert_mongo_upload(
                    db_src, src_entry_collection, upload_dict, upload_update, entry_update,
                    fix_problems, dataset_cache, stats, logger)
                if not dry:
                    _commit_upload(upload_dict, entry_dicts, dataset_dicts, doi_dicts, db_dst, stats)
                del entry_dicts, dataset_dicts  # To free up memory immediately
        except Exception as e:
            logger.error(f'Failed to migrate upload: {str(e)}', upload_id=upload_id, exc_info=e)
            count_failures += 1
            failed_and_skipped.append(upload_id)
        count_treated += 1
        # log progress report and estimated remaining time at regular intervals
        t = time.time()
        if t > next_report_time:
            elapsed = t - start_time
            progress = count_treated / number_of_uploads
            est_remain = elapsed * number_of_uploads / count_treated - elapsed
            print(
                f'Elapsed: {_seconds_to_h_m(elapsed)}, progress: {progress * 100.0:6.2f} %, '
                f'est. remain: {_seconds_to_h_m(est_remain)}')
            next_report_time += 60
    # Write failed ids to file, if requested
    if failed_ids_to_file:
        with open(failed_ids_to_file, 'w') as f:
            for upload_id in failed_and_skipped:
                f.write(upload_id + '\n')
    # Print a summary
    summary = f'Summary:\n\nElapsed time: {_seconds_to_h_m(time.time() - start_time)}\n\n'
    summary += f'{"Doc type":<10}{"found":>20}{"converted":>20}{"migrated":>20}\n'
    summary += '-' * 70 + '\n'
    for sub_stats in (stats.uploads, stats.entries, stats.datasets, stats.dois):
        summary += (
            f'{sub_stats.collection_name:<10}{sub_stats.total:>20}'  # pylint: disable=E1101
            f'{sub_stats.converted:>20}{sub_stats.migrated:>20}\n')
    if count_failures:
        summary += f'\n!!! THERE WERE ERRORS: {count_failures} upload(s) could not be converted/migrated\n\n'
    else:
        summary += '\nNo errors occurred :-)\n\n'
    if count_processing:
        summary += f'{count_processing} uploads were skipped since they are currently processing\n\n'
    if count_ignored:
        summary += f'{count_ignored} uploads were ignored because '
        if overwrite == 'if-newer':
            summary += 'they are not newer than the record in the destination db.\n\n'
        else:
            summary += 'they already exist in the destination db.\n\n'
    if dry:
        summary += 'Dry run - nothing written to the destination db\n\n'
    print(summary)
    if count_failures:
        return -1


def _convert_mongo_upload(
        db_src: Database, src_entry_collection: Collection, upload_dict: Dict[str, Any],
        upload_update: Dict[str, Any], entry_update: Dict[str, Any], fix_problems: bool,
        dataset_cache: Dict[str, _DatasetCacheItem], stats: _UpgradeStatistics, logger):
    '''
    Converts (upgrades) an upload_dict and all related records. If successful,
    returns three lists: one with converted entry dicts, and one with converted dataset dicts,
    and one with converted DOI dicts.
    '''
    upload_id = upload_dict['_id']
    published = upload_dict.get('publish_time') is not None
    # Determine which version we are migrating from
    is_v0 = 'user_id' in upload_dict

    _convert_mongo_proc(upload_dict)

    if is_v0:
        _rename_key(upload_dict, 'name', 'upload_name')
        _rename_key(upload_dict, 'create_time', 'upload_create_time')
        _rename_key(upload_dict, 'user_id', 'main_author')
        # Verify and then remove redundant legacy field 'published'.
        if 'published' in upload_dict:
            assert upload_dict['published'] == published, 'Inconsistency: published flag vs publish_time'
        # Set license if missing
        if 'license' not in upload_dict:
            upload_dict['license'] = 'CC BY 4.0'
        for key in _upload_keys_to_remove_v0:
            if key in upload_dict:
                upload_dict.pop(key)

    # Fetch all entries as a dictionary
    entry_dicts = list(src_entry_collection.find({'upload_id': upload_id}))
    stats.entries.total += len(entry_dicts)
    if is_v0 and entry_dicts:
        # Need to pass through all entries to check consistency and calculate common_coauthos
        first_metadata = entry_dicts[0].get('metadata', {})
        first_with_embargo = first_metadata.get('with_embargo')
        first_shared_with = first_metadata.get('shared_with')
        first_shared_with_set = set(first_shared_with or ())
        first_entry_uploader = first_metadata.get('uploader')
        first_external_db = first_metadata.get('external_db')
        first_entry_coauthors = first_metadata.get('coauthors', ())
        common_coauthors = set(_wrap_author(ca) for ca in first_entry_coauthors)

        fixed_external_db = False
        for entry_dict in entry_dicts:
            assert 'metadata' in entry_dict, 'Entry dict has no metadata key'
            entry_metadata_dict: Dict[str, Any] = entry_dict['metadata']
            with_embargo = entry_metadata_dict.get('with_embargo')
            assert with_embargo == first_with_embargo, 'Inconsistent embargo settings for entries'
            shared_with_set = set(entry_metadata_dict.get('shared_with') or ())
            assert shared_with_set == first_shared_with_set, 'Inconsistent shared_with settings for entries'
            uploader = entry_metadata_dict.get('uploader')
            assert uploader == first_entry_uploader, 'Inconsistent uploader for entries'
            external_db = entry_metadata_dict.get('external_db')
            if external_db != first_external_db:
                if external_db and first_external_db:
                    # Problem is unfixable (two different non-empty values encountered)
                    assert False, 'Inconsistent external_db for entries - unfixable'
                elif not fix_problems:
                    assert False, 'Inconsistent external_db for entries - use --fix-problems to fix'
                first_external_db = first_external_db or external_db  # Fix it
                fixed_external_db = True
            common_coauthors.intersection_update(
                _wrap_author(ca) for ca in entry_metadata_dict.get('coauthors', ()))

        if fixed_external_db:
            logger.warn('Fixed inconsistent external_db for upload', upload_id=upload_id)

        # Check embargo setting on the upload
        if published:
            if first_with_embargo:
                assert upload_dict.get('embargo_length'), (
                    f'Embargo flag set on entries, but no embargo_length specified on upload')
            else:
                upload_dict['embargo_length'] = 0
        if 'embargo_length' not in upload_dict:
            upload_dict['embargo_length'] = 0
        # Fields moving to the upload level
        upload_dict['reviewers'] = first_shared_with
        upload_dict['external_db'] = first_external_db
        # main_author
        if first_entry_uploader != upload_dict['main_author']:
            assert first_external_db, 'Different uploader on entry and upload, but external_db not set'
            # It's ok, but we should fetch main_author from entries.uploader instead of uplad.user_id
            upload_dict['main_author'] = first_entry_uploader
        # coauthors - put as many as possible in Upload.coauthors, the rest in entry.entry_coauthors
        if common_coauthors:
            upload_dict['coauthors'] = [ca for ca in first_entry_coauthors if _wrap_author(ca) in common_coauthors]
    else:
        common_coauthors = set(_wrap_author(ca) for ca in upload_dict.get('coauthors', ()))

    # Check that all required fields are there
    for field in ('_id', 'upload_create_time', 'main_author', 'embargo_length', 'license'):
        assert upload_dict.get(field) is not None, f'Missing required upload field: {field}'

    if upload_update:
        upload_dict.update(upload_update)

    # migrate entries
    newly_encountered_dataset_ids: Set[str] = set()
    for entry_dict in entry_dicts:
        assert not _is_processing(entry_dict), (
            f'the entry {entry_dict["_id"]} has status processing, but the upload is not processing.')
        _convert_mongo_entry(entry_dict, common_coauthors, fix_problems, logger)
        # Convert datasets
        datasets = entry_dict.get('datasets')
        if datasets:
            converted_datasets = []
            for dataset_id in datasets:
                if dataset_id in dataset_cache:
                    ds_cache = dataset_cache[dataset_id]
                else:
                    # First time we encounter this dataset ref
                    newly_encountered_dataset_ids.add(dataset_id)
                    ds_cache = _get_dataset_cache_data(dataset_id, db_src, logger)
                    dataset_cache[dataset_id] = ds_cache
                    if ds_cache.converted_dataset_dict is not None:
                        stats.datasets.total += 1
                    if ds_cache.converted_doi_dict is not None:
                        stats.dois.total += 1
                if ds_cache.converted_dataset_dict is None:
                    # Dataset record does not exist
                    assert fix_problems, (
                        f'Missing dataset reference encountered: {dataset_id} - '
                        'use --fix-problems to fix')
                    logger.warn(
                        'Fixing (removing) missing dataset reference for entry',
                        entry_id=entry_dict['_id'], dataset_id=dataset_id, upload_id=upload_id)
                elif not ds_cache.converted_dataset_dict:
                    # The value must be the empty dict, which represents a conversion failure
                    assert False, f'Reference to unconvertable dataset {dataset_id}'
                else:
                    # Dataset record exists and has been converted suuccessfully
                    converted_datasets.append(dataset_id)
            entry_dict['datasets'] = converted_datasets

        if entry_update:
            entry_dict.update(entry_update)

    # All conversion successful! Ready to migrate
    dataset_dicts: List[Dict[str, Any]] = []
    doi_dicts: List[Dict[str, Any]] = []
    for dataset_id in newly_encountered_dataset_ids:
        ds_cache = dataset_cache[dataset_id]
        if not ds_cache.ready_to_commit:
            if ds_cache.converted_dataset_dict:
                dataset_dicts.append(ds_cache.converted_dataset_dict)
                stats.datasets.converted += 1
            if ds_cache.converted_doi_dict:
                doi_dicts.append(ds_cache.converted_doi_dict)
                stats.dois.converted += 1
            ds_cache.ready_to_commit = True
    stats.entries.converted += len(entry_dicts)
    stats.uploads.converted += 1
    return entry_dicts, dataset_dicts, doi_dicts


def _convert_mongo_entry(entry_dict: Dict[str, Any], common_coauthors: Set, fix_problems: bool, logger):
    _convert_mongo_proc(entry_dict)
    # Validate the id and possibly fix problems
    generated_entry_id = utils.generate_entry_id(
        entry_dict['upload_id'], entry_dict['mainfile'], entry_dict.get('mainfile_key'))
    if entry_dict['_id'] != generated_entry_id:
        if not fix_problems:
            assert False, f'Entry id {entry_dict["_id"]} does not match generated value - use --fix-problems to fix'
        logger.warn('Fixing bad id for entry', old_id=entry_dict['_id'], entry_id=generated_entry_id)
        entry_dict['_id'] = generated_entry_id
    # Convert old v0 metadata
    if 'metadata' in entry_dict:
        _rename_key(entry_dict, 'calc_id', 'entry_id')
        _rename_key(entry_dict, 'parser', 'parser_name')
        _rename_key(entry_dict, 'create_time', 'entry_create_time')
        _rename_key(entry_dict, 'metadata.last_processing', 'last_processing_time')
        _rename_key(entry_dict, 'metadata.last_edit', 'last_edit_time')
        _rename_key(entry_dict, 'metadata.calc_hash', 'entry_hash')

        entry_metadata = entry_dict['metadata']
        # Entry coauthors
        entry_coauthors = entry_metadata.get('coauthors')
        if entry_coauthors:
            entry_coauthors = [ca for ca in entry_coauthors if _wrap_author(ca) not in common_coauthors]
            entry_dict['entry_coauthors'] = entry_coauthors or None
        # Flatten keys
        for key in _metadata_keys_to_flatten_v0:
            if key in entry_metadata:
                entry_dict[key] = entry_metadata.pop(key)
        # Remove redunant keys from the metadata dict; then it should be empty
        for key in _metadata_keys_to_remove_v0:
            if key in entry_metadata:
                entry_metadata.pop(key)
        assert not entry_metadata, f'Unexpected fields in Calc.metadata: {repr(entry_metadata)}'
        entry_dict.pop('metadata')

    # Check that all required fields are populated
    for field in ('_id', 'upload_id', 'entry_create_time', 'parser_name'):
        assert entry_dict.get(field) is not None, f'Missing required entry field: {field}'

    # Check if the parser exists
    parser_name = entry_dict.get('parser_name')
    if parser_name == 'parsers/gpaw2':
        parser_name = 'parsers/gpaw'
    entry_dict['parser_name'] = parser_name
    assert parser_name in parser_dict, f'Parser does not exist: {parser_name}'


def _convert_mongo_proc(proc_dict: Dict[str, Any]):
    if 'tasks_status' in proc_dict:
        # Old v0 version
        process_status = proc_dict['tasks_status']
        if process_status == 'CREATED':
            process_status = ProcessStatus.READY
        assert process_status in ProcessStatus.STATUSES_VALID_IN_DB, f'Invalid tasks_status: {process_status}'
        proc_dict['process_status'] = process_status
        last_status_message = proc_dict.get('last_status_message')
        if not last_status_message:
            # Generate a nicer last_status_message
            current_process: str = proc_dict.get('current_process')
            errors: List[str] = proc_dict.get('errors')
            if errors:
                last_status_message = f'Process {current_process} failed: {errors[-1]}'
            elif current_process and process_status == ProcessStatus.SUCCESS:
                last_status_message = f'Process {current_process} completed successfully'
            else:
                assert False, f'Unexpected proc state: {current_process} / {process_status}'
            proc_dict['last_status_message'] = last_status_message

        for field in ('tasks', 'current_task', 'tasks_status'):
            proc_dict.pop(field, None)


def _convert_mongo_dataset(dataset_dict: Dict[str, Any]):
    _rename_key(dataset_dict, 'name', 'dataset_name')
    _rename_key(dataset_dict, 'created', 'dataset_create_time')
    _rename_key(dataset_dict, 'modified', 'dataset_modified_time')
    # Check that all required fields are there
    for field in ('dataset_name',):
        assert dataset_dict.get(field) is not None, f'Dataset missing required field {field}'


def _convert_mongo_doi(doi_dict: Dict[str, Any]):
    pass


def _get_dataset_cache_data(dataset_id: str, db_src: Database, logger) -> _DatasetCacheItem:
    '''
    Fetches and converts a dataset and related doi record. None values will be used in the
    returned :class:`_DatasetCacheItem` if a record is missing; empty dicts are used if the
    record exists but can't be converted
    '''
    dataset_dict = db_src.dataset.find_one({'_id': dataset_id})
    doi_dict = None
    if dataset_dict:
        try:
            _convert_mongo_dataset(dataset_dict)
            doi = dataset_dict.get('doi')
            if doi:
                doi_dict = db_src.d_o_i.find_one({'_id': doi})
                # DOI records were not stored originally, so some older datasets with
                # doi might not have a corresponding DOI record. The data is also
                # not super critical, so if it doesn't exist we will just ignore it
                if doi_dict:
                    _convert_mongo_doi(doi_dict)
        except Exception as e:
            # Conversion failed
            logger.error(str(e), dataset_id=dataset_id, exc_info=e)
            dataset_dict = {}
            doi_dict = {}
    return _DatasetCacheItem(converted_dataset_dict=dataset_dict, converted_doi_dict=doi_dict)


def _is_processing(proc_dict: Dict[str, Any]) -> bool:
    process_status = proc_dict.get('tasks_status')  # Used in v0
    if not process_status:
        process_status = proc_dict['process_status']
    return process_status in ProcessStatus.STATUSES_PROCESSING


def _commit_upload(
        upload_dict: Dict[str, Any], entry_dicts: List[Dict[str, Any]],
        dataset_dicts: List[Dict[str, Any]], doi_dicts: List[Dict[str, Any]],
        db_dst: Database, stats: _UpgradeStatistics):
    # Commit datasets
    if dataset_dicts:
        dataset_writes = []
        for dataset_dict in dataset_dicts:
            dataset_writes.append(ReplaceOne({'_id': dataset_dict['_id']}, dataset_dict, upsert=True))
        db_dst.dataset.bulk_write(dataset_writes)
        stats.datasets.migrated += len(dataset_dicts)
    # Commit DOIs
    if doi_dicts:
        doi_writes = []
        for doi_dict in doi_dicts:
            doi_writes.append(ReplaceOne({'_id': doi_dict['_id']}, doi_dict, upsert=True))
        db_dst.d_o_i.bulk_write(doi_writes)
        stats.dois.migrated += len(doi_writes)
    # Commit upload
    db_dst.upload.replace_one({'_id': upload_dict['_id']}, upload_dict, upsert=True)
    stats.uploads.migrated += 1
    # Commit entries
    if entry_dicts:
        entry_writes = []
        for entry_dict in entry_dicts:
            entry_writes.append(ReplaceOne({'_id': entry_dict['_id']}, entry_dict, upsert=True))
        db_dst.entry.bulk_write(entry_writes)
        stats.entries.migrated += len(entry_dicts)


def _rename_key(d: Dict[str, Any], old_name: str, new_name: str):
    '''
    Renames a key in the provided dictionary `d`, from `old_name` to `new_name`. We may use
    "point notation" in `old_name`, i.e. "metadata.external_id" will look for a
    key 'external_id' in d['metadata'].
    '''
    parts = old_name.split('.')
    d_current = d
    for i, part in enumerate(parts):
        if i == len(parts) - 1:
            if d_current and part in d_current:
                d[new_name] = d_current.pop(part)
        else:
            if d_current and part in d_current:
                d_current = d_current[part]
            else:
                break  # Not found


def _seconds_to_h_m(seconds: float):
    ''' Converst a time in seconds to a nicer string with hour and minute parts. '''
    return f'{seconds // 3600:>3.0f} h {(seconds % 3600) / 60:2.0f} min'


def _wrap_author(author):
    '''
    Takes an author reference (str or dict) and if it is a dict, converts it into a tuple.
    If the author is a str, it is instead returned as it is. This is used to get an object
    which is hashable and can be used in sets.
    '''
    if type(author) == str:
        return author
    return tuple((k, author[k]) for k in sorted(author.keys()))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Union, Dict, Any
from datetime import datetime, timedelta

from nomad import search, files
from nomad.datamodel import EntryMetadata, EntryArchive, Results
from nomad.datamodel.metainfo.workflow import Workflow
from nomad.datamodel.metainfo.simulation.run import Run, Program
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from nomad.processing.data import _mongo_upload_metadata
from nomad.normalizing import normalizers


class ExampleData:
    '''
    Allows to define, create, and manage a set of example data. Will create respective
    data via raw files, archives, in mongodb, and in both elasticsearch indices.

    Requires initialized infrastructure.

    Attributes:
        uploads: A dictionary with with upload_ids as keys and lists of entry_ids as values.
        entries: A dictionary with entry_ids as keys and their ``EntryMetadata`` as values.
        archives: A dictionary with entry_ids as keys and their ``EntryArchives`` as values.
    '''

    def __init__(self, **kwargs):
        self.upload_entries: Dict[str, List[str]] = dict()
        self.uploads: Dict[str, Dict[str, Any]] = dict()
        self.entries: Dict[str, EntryMetadata] = dict()
        self.archives: Dict[str, EntryArchive] = dict()

        self.entry_defaults = kwargs
        self._entry_id_counter = 1

        self._time_stamp = datetime.utcnow()

    def save(self, with_files: bool = True, with_mongo: bool = True, with_es: bool = True):
        from tests.test_files import create_test_upload_files
        from nomad import processing as proc

        # Save
        if with_mongo:
            for upload_id, upload_dict in self.uploads.items():
                upload_dict['main_author'] = upload_dict['main_author'].user_id
                mongo_upload = proc.Upload(**upload_dict)
                mongo_upload.save()

            for entry_metadata in self.entries.values():
                process_status = (
                    proc.ProcessStatus.SUCCESS if entry_metadata.processed else proc.ProcessStatus.FAILURE)
                mongo_entry = proc.Entry(
                    entry_create_time=entry_metadata.entry_create_time,
                    entry_id=entry_metadata.entry_id,
                    upload_id=entry_metadata.upload_id,
                    mainfile=entry_metadata.mainfile,
                    parser_name='parsers/vasp',
                    process_status=process_status)
                mongo_entry.set_mongo_entry_metadata(entry_metadata)
                mongo_entry.save()

        if with_es:
            archives = list(self.archives.values())
            search.index(archives, update_materials=True, refresh=True)

        if with_files:
            for upload_id, upload_dict in self.uploads.items():
                entry_ids = self.upload_entries.get(upload_id, [])
                archives = []
                for entry_id in entry_ids:
                    if entry_id in self.archives:
                        archives.append(self.archives[entry_id])

                create_test_upload_files(
                    upload_id, archives, published=upload_dict.get('publish_time') is not None,
                    embargo_length=upload_dict['embargo_length'])
                from nomad import files
                assert files.UploadFiles.get(upload_id) is not None

    def delete(self):
        from nomad import processing as proc

        for upload_id in self.upload_entries:
            search.delete_upload(upload_id, refresh=True)
            upload_proc = proc.Upload.objects(upload_id=upload_id).first()
            if upload_proc is not None:
                upload_proc.delete()
            upload_files = files.UploadFiles.get(upload_id)
            if upload_files is not None:
                upload_files.delete()

    def create_upload(self, upload_id, published=None, **kwargs):
        '''
        Creates a dictionary holding all the upload information.
        Default values are used/generated, and can be set via kwargs.
        '''
        upload_dict = {
            'upload_id': upload_id,
            'current_process': 'process_upload',
            'process_status': 'SUCCESS',
            'errors': [],
            'warnings': [],
            'upload_create_time': self._next_time_stamp(),
            'complete_time': self._next_time_stamp(),
            'last_update': self._next_time_stamp(),
            'embargo_length': 0,
            'publish_time': None,
            'license': 'CC BY 4.0',
            'published_to': []}
        upload_dict.update(kwargs)
        if published is not None:
            if published and not upload_dict['publish_time']:
                upload_dict['publish_time'] = self._next_time_stamp()
            elif not published:
                assert not upload_dict.get('publish_time')
        if 'main_author' not in upload_dict and 'main_author' in self.entry_defaults:
            upload_dict['main_author'] = self.entry_defaults['main_author']
        if 'coauthors' not in upload_dict and 'coauthors' in self.entry_defaults:
            upload_dict['coauthors'] = self.entry_defaults['coauthors']
        if 'reviewers' not in upload_dict and 'reviewers' in self.entry_defaults:
            upload_dict['reviewers'] = self.entry_defaults['reviewers']
        self.uploads[upload_id] = upload_dict

    def create_entry(
            self,
            entry_archive: EntryArchive = None,
            entry_id: str = None, upload_id: str = None,
            material_id: str = None,
            mainfile: str = None,
            results: Union[Results, dict] = None,
            archive: dict = None, **kwargs) -> EntryArchive:

        assert upload_id in self.uploads, 'Must create the upload first'
        upload_dict = self.uploads[upload_id]

        if entry_id is None:
            entry_id = f'test_entry_id_{self._entry_id_counter}'
            self._entry_id_counter += 1

        if mainfile is None:
            mainfile = f'mainfile_for_{entry_id}'

        if entry_archive is None:
            entry_archive = EntryArchive()

        if material_id is None:
            material_id = 'test_material_id'

        entry_metadata = entry_archive.metadata
        if entry_metadata is None:
            entry_metadata = entry_archive.m_create(EntryMetadata)

        entry_metadata.m_update(
            entry_id=entry_id,
            upload_id=upload_id,
            mainfile=mainfile,
            entry_hash='dummy_hash_' + entry_id,
            domain='dft',
            entry_create_time=self._next_time_stamp(),
            processed=True,
            parser_name='parsers/vasp')
        entry_metadata.m_update(**self.entry_defaults)
        # Fetch data from Upload
        upload_values = {k: upload_dict[k] for k in _mongo_upload_metadata if k in upload_dict}
        upload_values['with_embargo'] = upload_dict['embargo_length'] > 0
        upload_values['published'] = upload_dict.get('publish_time') is not None
        for k in list(_mongo_upload_metadata) + ['with_embargo', 'published']:
            assert k not in kwargs, f'Upload level metadata specified on entry level: {k}'
        entry_metadata.m_update(**upload_values)
        entry_metadata.m_update(**kwargs)

        # create v1 default data
        if entry_archive.results is None:
            if results is None:
                results = {
                    'material': {
                        'material_id': material_id,
                        'elements': ['H', 'O'],
                        'nelements': 2,
                        'symmetry': {
                            'crystal_system': 'cubic'
                        }
                    },
                    'method': {
                        'simulation': {
                            'program_name': 'VASP',
                            'dft': {
                                'xc_functional_type': 'GGA'
                            }
                        }
                    },
                    'properties': {
                        'n_calculations': 1,
                        'electronic': {
                            'dos_electronic': {
                                'spin_polarized': entry_id.endswith('04'),
                                'band_gap': [
                                    {
                                        'type': 'direct' if entry_id.endswith('04') else 'indirect'
                                    }
                                ]
                            }
                        }
                    }
                }
            if isinstance(results, dict):
                section_results = Results.m_from_dict(results)
            else:
                section_results = results
            assert isinstance(section_results, Results)
            entry_archive.m_add_sub_section(EntryArchive.results, section_results)

        if len(entry_archive.run) == 0:
            entry_archive.m_create(Run)

        if archive is not None:
            entry_archive.m_update(**archive)

        if entry_archive.results.material.material_id is None:
            entry_archive.results.material.material_id = material_id

        self.archives[entry_id] = entry_archive
        self.entries[entry_id] = entry_metadata
        self.upload_entries.setdefault(entry_metadata.upload_id, []).append(entry_id)

        return entry_archive

    def _next_time_stamp(self):
        '''
        Returns self._time_stamp and ticks up the time stamp with 1 millisecond. This
        utility guarantees that we get unique and increasing time stamps for each entity.
        '''
        self._time_stamp += timedelta(milliseconds=1)
        return self._time_stamp

    def create_structure(
            self,
            upload_id: str, id: int, h: int, o: int, extra: List[str], periodicity: int,
            optimade: bool = True, metadata: dict = None):

        ''' Creates an entry in Elastic and Mongodb with the given properties.

        Does require initialized :func:`elastic_infra` and :func:`mongo_infra`.

        Args:
            meta_info: A legace metainfo env.
            id: A number to create ``test_entry_id_<number>`` ids.
            h: The amount of H atoms
            o: The amount of O atoms
            extra: A list of further atoms
            periodicity: The number of dimensions to repeat the structure in
            optimade: A boolean. If true the entry will have optimade metadata. Default is True.
            metadata: Additional (user) metadata.
        '''
        test_vector = [0, 0, 0]
        atom_labels = ['H' for i in range(0, h)] + ['O' for i in range(0, o)] + extra

        archive = EntryArchive()
        run = archive.m_create(Run)
        run.m_create(Program, name='VASP')
        run.m_create(System).m_create(
            Atoms,
            labels=atom_labels,
            positions=[test_vector for i in range(0, len(atom_labels))],
            lattice_vectors=[test_vector, test_vector, test_vector],
            periodic=[True for _ in range(0, periodicity)] + [False for _ in range(periodicity, 3)])

        for normalizer_class in normalizers:
            normalizer = normalizer_class(archive)
            normalizer.normalize()

        entry_metadata = archive.metadata
        entry_metadata.domain = 'dft'
        entry_metadata.apply_archvie_metadata(archive)

        if not optimade:
            entry_metadata.optimade = None
            entry_metadata.quantities.remove('metadata.optimade')

        if metadata is not None:
            kwargs = metadata
        else:
            kwargs = {}

        self.create_entry(
            entry_archive=archive,
            upload_id=upload_id, entry_id='test_entry_id_%d' % id, domain='dft', **kwargs)


def create_entry_archive(metadata, results: dict = None, run: dict = None, workflows: List = None):
    '''Creates an entry archive out of python objects.

    Args:
        metadata: The archive metadata
        results: The archive results
        run: The archive run
        workflows: List of archive workflows
    '''
    entry = EntryArchive()
    if metadata:
        entry_metadata = entry.m_create(EntryMetadata)
        entry_metadata.m_update(**metadata)
    if results:
        entry_results = Results.m_from_dict(results)
        entry.m_add_sub_section(EntryArchive.results, entry_results)
    if run:
        entry_run = Run.m_from_dict(run)
        entry.m_add_sub_section(EntryArchive.run, entry_run)
    if workflows:
        for workflow in workflows:
            entry_workflow = Workflow.m_from_dict(workflow)
            entry.m_add_sub_section(EntryArchive.workflow, entry_workflow)

    return entry



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A Template object to control and separate template paths according to optionality"""

import copy
import json

from nexusparser.tools.dataconverter import helpers


class Template(dict):
    """A Template object to control and separate template paths according to optionality"""

    def __init__(self, template=None, **kwargs):
        super(Template, self).__init__(**kwargs)
        if template is None:
            self.optional: dict = {}
            self.recommended: dict = {}
            self.required: dict = {}
            self.undocumented: dict = {}
            self.optional_parents: list = []
        else:
            self.optional: dict = copy.deepcopy(template["optional"])
            self.recommended: dict = copy.deepcopy(template["recommended"])
            self.required: dict = copy.deepcopy(template["required"])
            self.undocumented: dict = copy.deepcopy(template["undocumented"])
            self.optional_parents: list = copy.deepcopy(template["optional_parents"])

    def get_accumulated_dict(self):
        """Returns a dictionary of all the optionalities merged into one."""
        return {**self.optional, **self.recommended, **self.required, **self.undocumented}

    def __repr__(self):
        """Returns a unique string representation for the Template object."""
        return self.get_accumulated_dict().__repr__()

    def __str__(self):
        """Returns a readable string representation for the Template object."""
        accumulated_dict = self.get_accumulated_dict()
        accumulated_dict.update((key, "None") for key in accumulated_dict)
        return json.dumps(accumulated_dict, indent=4, sort_keys=True)

    def __setitem__(self, k, v):
        """Handles how values are set within the Template object."""
        if k.startswith("/"):
            if k in self.recommended:
                self.recommended[k] = v
            elif k in self.required:
                self.required[k] = v
            elif k in self.optional:
                self.optional[k] = v
            else:
                self.undocumented[k] = v
        else:
            raise KeyError("You cannot add non paths to the root template object. "
                           "Place them appropriately e.g. template[\"optional\"]"
                           "[\"/ENTRY[entry]/data/path\"]")

    def keys(self):
        """Returns the list of keys stored in the Template object."""
        return list(self.optional.keys()) + \
            list(self.recommended.keys()) + \
            list(self.required.keys()) + \
            list(self.undocumented.keys())

    def items(self):
        """Returns a list of tuples of key, value stored in the Template object."""
        return self.get_accumulated_dict().items()

    def __iter__(self):
        return dict.__iter__(self.get_accumulated_dict())

    def get_optionality(self, optionality):
        """Returns the dictionary for given optionality"""
        if optionality == "optional":
            return self.optional
        if optionality == "recommended":
            return self.recommended
        if optionality == "required":
            return self.required
        if optionality == "undocumented":
            return self.undocumented
        return self.required

    def get_documented(self):
        """Returns a dictionary of all the optionalities merged into one."""
        return {**self.optional, **self.recommended, **self.required}

    def __getitem__(self, k):
        """Handles how values are accessed from the Template object."""
        # Try setting item in all else throw error. Does not append to default.
        if k == "optional_parents":
            return self.optional_parents
        if k.startswith("/"):
            try:
                return self.optional[k]
            except KeyError:
                try:
                    return self.recommended[k]
                except KeyError:
                    try:
                        return self.required[k]
                    except KeyError:
                        return self.undocumented[k]
        return self.get_optionality(k)

    def clear(self):
        """Clears all data stored in the Template object."""
        for del_dict in (self.optional, self.recommended, self.required, self.undocumented):
            del_dict.clear()

    def rename_entry(self, old_name: str, new_name: str, deepcopy=True):
        """Rename all entries under old name to new name."""
        for internal_dict in (self.optional, self.recommended, self.required, self.undocumented):
            keys = list(internal_dict.keys())
            for key in keys:
                entry_name = helpers.get_name_from_data_dict_entry(key.split("/")[1])

                entry_search_term = f"{entry_name}]"
                rest_of_path = key[key.index(entry_search_term) + len(entry_search_term):]
                if entry_name == old_name:
                    value = internal_dict[key] if deepcopy else None
                    internal_dict[f"/ENTRY[{new_name}]{rest_of_path}"] = value
                    del internal_dict[key]

    def update(self, template):
        """Merges second template to original"""
        for optionality in ("optional", "recommended", "required", "undocumented"):
            self.get_optionality(optionality).update(template.get_optionality(optionality))

    def add_entry(self, entry_name):
        """Add the whole NXDL again with a new HDF5 name for the template."""
        template = Template(self)
        template.rename_entry("entry", entry_name, False)
        self.update(template)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import os
import numpy as np
import fnmatch

try:
    import MDAnalysis
except Exception:
    logging.warn('Error importing required MDAnalysis.')
    MDAnalysis = None
from nomad.parsing.file_parser.text_parser import Quantity, TextParser, FileParser
from nomad.units import ureg
from nomad.datamodel.metainfo.simulation.run import Run, Program
from nomad.datamodel.metainfo.simulation.system import (
    System, Atoms
)
from nomad.datamodel.metainfo.simulation.method import (
    Method, ForceField, Model, AtomParameters
)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Energy, EnergyEntry, Forces, ForcesEntry, Thermodynamics, Vibrations
)
from nomad.datamodel.metainfo.workflow import (
    Workflow, GeometryOptimization, MolecularDynamics
)
from .metainfo.tinker import x_tinker_section_control_parameters

re_n = r'[\n\r]'
re_f = r'[-+]?\d+\.\d*(?:[DdEe][-+]\d+)?'
mol = (1 * ureg.mol).to('particle').magnitude


class TrajParser(FileParser):
    def __init__(self):
        super().__init__()

    @property
    def universe(self):
        if self._file_handler is None and self.mainfile is not None:
            try:
                self._results = dict()
                topology_format = 'ARC' if self.mainfile.endswith('.arc') else 'TXYZ'
                self._file_handler = MDAnalysis.Universe(self.mainfile, topology_format=topology_format)
            except Exception:
                self.logger.error('Error reading tinker trajectory file.')

        return self._file_handler


class KeyParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def to_key_val(val_in):
            val = val_in.strip().split()
            return [val[0], True] if len(val) == 1 else val

        self._quantities = [Quantity(
            'key_val',
            rf'([a-z\-]+) *([^#]*?) *{re_n}',
            str_operation=to_key_val, repeats=True)]


class RunParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def to_argument(val_in):
            val_in = val_in.strip().split()
            argument = dict()
            if '-k' in val_in:
                argument['key'] = val_in.pop(val_in.index('-k') + 1)
                val_in.remove('-k')
            argument['name'] = val_in.pop(0)
            argument['parameters'] = val_in
            return argument

        self._quantities = [
            Quantity(
                'molecular_dynamics',
                r'dynamic +(\S+ +\-*k* *\S+.+)',
                repeats=True, str_operation=to_argument
            ),
            Quantity(
                'geometry_optimization',
                r'minimize +(\S+ +\-*k* *\S+.+)',
                repeats=True, str_operation=to_argument
            ),
            Quantity(
                'single_point',
                r'vibrate +(\S+ +\-*k* *\S+.+)',
                repeats=True, str_operation=to_argument
            )
        ]


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        iteration_quantity = Quantity(
            'iteration',
            rf'Iter.+([\s\S]+?){re_n} *{re_n}',
            sub_parser=TextParser(quantities=[Quantity(
                'step',
                rf' +\d+ +({re_f} +{re_f} +[\d\.DE\+\- ]+)',
                str_operation=lambda x: [float(v) for v in x.replace('D+', 'E+').replace('d+', 'e+').split()],
                repeats=True, dtype=np.dtype(np.float64)
            )])
        )

        calculation_quantities = [
            Quantity('program_version', r'Version ([\d\.]+)', dtype=str),
            Quantity(
                'vibrate',
                r'(Eigenvalues of the Hessian Matrix[\s\S]+?)\Z',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'eigenvalues',
                        r'Eigenvalues of the Hessian Matrix :\s+([\d\-\s\.]+)',
                        dtype=np.dtype(np.float64)
                    ),
                    Quantity(
                        'frequencies',
                        r'Vibrational Frequencies \(cm\-1\) :([\d\-\s\.]+)',
                        dtype=np.dtype(np.float64)
                    ),
                ])),
            Quantity(
                'minimize',
                r'(.+?Optimization :[\s\S]+?Final Gradient Norm.+)',
                sub_parser=TextParser(quantities=[
                    iteration_quantity,
                    Quantity('method', r' +(.+) Optimization', flatten=False, dtype=str),
                    Quantity('x_tiner_final_function_value', rf'Final Function Value : +({re_f})', dtype=np.float64),
                    Quantity('x_tinker_final_rms_gradient', rf'Final RMS Gradient : +({re_f})', dtype=np.float64),
                    Quantity('x_tinker_final_gradient_norm', rf'Final Gradient Norm : +({re_f})', dtype=np.float64)
                ])
            ),
            Quantity(
                'dynamic',
                r'(Molecular Dynamics[\s\S]+?)(?:\#\#\#\#\#|\Z)',
                sub_parser=TextParser(quantities=[
                    iteration_quantity,
                    Quantity(
                        'instantaneous_values',
                        r'(Instantaneous Values for Frame Saved at[\s\S]+?Coordinate File.+)',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity('step', r'(\d+) Dynamics Steps', dtype=np.int32),
                            Quantity(
                                'time',
                                rf'Current Time +({re_f}) Picosecond',
                                dtype=np.float64, unit=ureg.ps
                            ),
                            Quantity(
                                'potential',
                                rf'Current Potential +({re_f}) Kcal/mole',
                                dtype=np.float64, unit=ureg.kcal / mol
                            ),
                            Quantity(
                                'kinetic',
                                rf'Current Kinetic +({re_f}) Kcal/mole',
                                dtype=np.float64, unit=ureg.kcal / mol
                            ),
                            Quantity(
                                'lattice_lengths',
                                rf'Lattice Lengths +({re_f} +{re_f} +{re_f})',
                                dtype=np.dtype(np.float64)
                            ),
                            Quantity(
                                'lattice_angles',
                                rf'Lattice Angles +({re_f} +{re_f} +{re_f})',
                                dtype=np.dtype(np.float64)
                            ),
                            Quantity('frame', r'Frame Number +(\d+)', dtype=np.int32),
                            Quantity('coordinate_file', r'Coordinate File +(\S+)', dtype=str)
                        ])
                    ),
                    Quantity(
                        'average_values',
                        r'(Average Values for the Last[\s\S]+?Density.+)',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity('step', r'(\d+) Dynamics Steps', dtype=np.int32),
                            Quantity(
                                'time',
                                rf'Simulation Time +({re_f}) Picosecond',
                                dtype=np.float64, unit=ureg.ps
                            ),
                            Quantity(
                                'energy_total',
                                rf'Total Energy +({re_f}) Kcal/mole',
                                dtype=np.float64, unit=ureg.kcal / mol
                            ),
                            Quantity(
                                'potential',
                                rf'Potential Energy +({re_f}) Kcal/mole',
                                dtype=np.float64, unit=ureg.kcal / mol
                            ),
                            Quantity(
                                'kinetic',
                                rf'Kinetic Energy +({re_f}) Kcal/mole',
                                dtype=np.float64, unit=ureg.kcal / mol
                            ),
                            Quantity(
                                'temperature',
                                rf'Temperature +({re_f}) Kelvin',
                                dtype=np.float64, unit=ureg.kelvin
                            ),
                            Quantity(
                                'pressure',
                                rf'Pressure +({re_f}) Atmosphere',
                                dtype=np.float64, unit=ureg.atm
                            ),
                            Quantity(
                                'density',
                                rf'Density +({re_f}) Grams/cc',
                                dtype=np.float64, unit=ureg.g / ureg.cm ** 3
                            ),

                        ])
                    )
                ])
            )
        ]

        self._quantities = [
            Quantity(
                'run',
                r'Software Tools for Molecular Design([\s\S]+?)(?:TINKER  \-\-\-|\Z)',
                repeats=True, sub_parser=TextParser(quantities=calculation_quantities)
            )
        ]


class TinkerParser:
    def __init__(self):
        self.out_parser = OutParser()
        self.traj_parser = TrajParser()
        self.key_parser = KeyParser()
        self.run_parser = RunParser()
        self._run_types = {
            'vibrate': 'single_point', 'minimize': 'geometry_optimization',
            'dynamic': 'molecular_dynamics'}

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger
        self._base_name = os.path.basename(self.filepath).rsplit(".", 1)[0]
        files = os.listdir(self.maindir)
        # required to track the current files
        self._files = dict()
        for ext in ['xyz', 'arc', 'key']:
            matches = fnmatch.filter(files, f'{self._base_name}.{ext}*')
            if not matches:
                matches = fnmatch.filter(files, f'tinker.{ext}*')
            matches.sort(key=lambda x: int(x.rsplit('_', 1)[-1]) if x[-1].isdecimal() else 0)
            self._files[ext] = dict(files=matches, current=matches[0] if matches else '')

    def _get_tinker_file(self, ext):
        if ext not in self._files:
            return
        current = self._files[ext]['current']
        # advance to the next file
        files = self._files[ext]['files']
        try:
            self._files[ext]['current'] = files[files.index(current) + 1]
        except Exception:
            # the last file has been reached in this case
            pass
        return os.path.join(self.maindir, current)

    def parse_system(self, index, filename):
        self.traj_parser.mainfile = filename
        if self.traj_parser.universe is None:
            return

        sec_system = self.archive.run[-1].m_create(System)
        trajectory = self.traj_parser.universe.trajectory[index]
        sec_system.atoms = Atoms(
            positions=trajectory.positions * ureg.angstrom,
            labels=[atom.name for atom in list(self.traj_parser.universe.atoms)])
        if trajectory.triclinic_dimensions is not None:
            sec_system.atoms.lattice_vectors = trajectory.triclinic_dimensions * ureg.angstrom
        if trajectory.has_velocities:
            sec_system.atoms.velocities = trajectory.velocities * (ureg.angstrom / ureg.ps)

        return sec_system

    def parse_method(self):
        sec_method = self.archive.run[-1].m_create(Method)

        parameters = self.archive.run[-1].x_tinker_control_parameters
        if parameters.get('parameters') is not None:
            sec_method.force_field = ForceField(model=[Model(name=parameters['parameters'])])
            # TODO read the prm file

        property_map = {
            'name': 'label', 'type': 'x_tinker_atom_type', 'resid': 'x_tinker_atom_resid'
        }
        if self.traj_parser.universe is not None:
            for atom in list(self.traj_parser.universe.atoms):
                sec_atom = sec_method.m_create(AtomParameters)
                for key in ['charge', 'mass', 'name', 'type', 'resid']:
                    if hasattr(atom, key):
                        setattr(sec_atom, property_map.get(key, key), getattr(atom, key))

        # TODO add interaction parameters

    def parse_workflow(self, program, run):
        def resolve_ensemble_type():
            parameters = self.archive.run[-1].x_tinker_control_parameters
            if parameters is None:
                return

            thermostat, barostat = parameters.get('thermostat', ''), parameters.get('barostat', '')
            thermostats = ['berendsen', 'andersen', 'bussi']
            # TODO verify this
            if barostat.lower() in ['berendsen', 'montecarlo']:
                ensemble_type = 'NPT'
            elif not barostat and thermostat in thermostats:
                ensemble_type = 'NVT'
            elif not barostat and not thermostat:
                ensemble_type = 'NVE'
            else:
                ensemble_type = None

            return ensemble_type

        workflow_type = self.archive.workflow[-1].type
        parameters = list(program.get('parameters', []))
        # so we cover the case when optional parameters are missing
        parameters.extend([None] * 6)
        if workflow_type == 'molecular_dynamics':
            sec_md = self.archive.workflow[-1].m_create(MolecularDynamics)
            control_parameters = self.archive.run[-1].x_tinker_control_parameters
            # TODO verify this! I am sure it is wrong but tinker documentation does not specify clearly
            ensemble_types = ['NVE', 'NVT', 'NPT', None, None]
            sec_md.ensemble_type = ensemble_types[int(parameters[3]) - 1] if parameters[3] is not None else resolve_ensemble_type()
            sec_md.timestep = parameters[1] * ureg.fs
            sec_md.x_tinker_barostat_tau = control_parameters.get('tau-pressure')
            sec_md.x_tinker_barostat_type = control_parameters.get('barostat')
            sec_md.x_tinker_integrator_type = control_parameters.get('integrator')
            sec_md.x_tinker_number_of_steps_requested = parameters[0]
            sec_md.x_tinker_integrator_dt = parameters[1] * ureg.ps
            sec_md.x_tinker_thermostat_target_temperature = parameters[4] * ureg.kelvin
            sec_md.x_tinker_barostat_target_pressure = parameters[5] * ureg.atmosphere
            sec_md.x_tinker_thermostat_tau = control_parameters.get('tau-temperature')
            sec_md.x_tinker_thermostat_type = control_parameters.get('thermostat')

        elif workflow_type == 'geometry_optimization':
            sec_opt = self.archive.workflow[-1].m_create(GeometryOptimization)
            sec_opt.method = run.minimize.get('method')
            sec_opt.x_tinker_convergence_tolerance_rms_gradient = parameters[0]
            for key, val in run.minimize.items():
                setattr(sec_opt, key, val)

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.maindir = os.path.dirname(self.filepath)
        self.init_parser()

        def resolve_workflow_type(run):
            for program in run.keys():
                if program in self._run_types and run.get(program) is not None:
                    return self._run_types[program]

        def get_reference_filename(program):
            # resolves the filename as provided in the cli command for the program
            filename = program.get('name', '')
            if '.xyz' not in filename:
                filename = self._files['xyz']['current']
            # succeding files start from the index of the file for geometry_opt and md
            filename = filename if '.xyz' in filename else f'{filename}.xyz'
            if self.archive.workflow[-1].type in ['geometry_optimization', 'molecular_dynamics']:
                files = self._files['xyz']['files']
                try:
                    self._files['xyz']['current'] = files[files.index(filename) + 1]
                except Exception:
                    pass
            return os.path.join(self.maindir, filename)

        # necesarry to extract program parameters from the cli command in basename.run
        self.run_parser.mainfile = os.path.join(self.maindir, f'{self._base_name}.run')

        for run in self.out_parser.get('run', []):
            sec_run = archive.m_create(Run)
            sec_run.program = Program(name='tinker', version=run.get('program_version'))

            sec_workflow = self.archive.m_create(Workflow)
            sec_workflow.type = resolve_workflow_type(run)

            # get parameters of the program from the cli command, the key file and the
            # initial structure file can also be specied as an argument so we need to
            # extract the information here
            # program can be executed for an arbitrary number of times so we need to
            # resolve the index of the appropriate command
            n_workflow = len([workflow for workflow in self.archive.workflow if workflow.type == sec_workflow.type]) - 1
            program = self.run_parser.get(sec_workflow.type, [])
            program = program[n_workflow] if len(program) > n_workflow else dict()
            if run.vibrate is not None:
                # reference structure
                sec_system = self.parse_system(0, get_reference_filename(program))
                sec_scc = sec_run.m_create(Calculation)
                sec_vibrations = sec_scc.m_create(Vibrations)
                sec_vibrations.value = [run.vibrate.frequencies[n] for n in range(len(
                    run.vibrate.get('frequencies', []))) if n % 2 == 1] * (1 / ureg.cm)
                sec_vibrations.x_tinker_eigenvalues = [run.vibrate.eigenvalues[n] for n in range(len(
                    run.vibrate.get('eigenvalues', []))) if n % 2 == 1]
                sec_scc.system_ref = sec_system

            if run.minimize is not None:
                # initial structure
                initial_system = self.parse_system(0, get_reference_filename(program))

                # optimized structure
                sec_system = self.parse_system(0, self._get_tinker_file('xyz'))

                for n, step in enumerate(run.minimize.get('iteration', {}).get('step', [])):
                    sec_scc = sec_run.m_create(Calculation)
                    sec_scc.energy = Energy(total=EnergyEntry(
                        value=step[0] * len(sec_system.atoms.positions) * ureg.kcal / mol))
                    if n == 0:
                        sec_scc.system_ref = initial_system
                # only the optimized structure is printed, corresponding to the last scc
                sec_scc.system_ref = sec_system

            if run.dynamic is not None:
                self.traj_parser.mainfile = self._get_tinker_file('arc')
                average_values = {value.step: value for value in run.dynamic.get('average_values', [])}
                for nframe, value in enumerate(run.dynamic.get('instantaneous_values', [])):
                    filename = os.path.join(self.maindir, value.get('coordinate_file', ''))

                    index = nframe if filename.endswith('.arc') else 0
                    sec_system = self.parse_system(index, filename)
                    n_atoms = len(sec_system.atoms.positions)
                    sec_scc = sec_run.m_create(Calculation)
                    sec_scc.energy = Energy(total=EnergyEntry(
                        value=(value.potential + value.kinetic) * n_atoms))
                    sec_scc.thermodynamics.append(Thermodynamics(
                        kinetic_energy=value.kinetic * n_atoms, potential_energy=value.potential * n_atoms,
                        time_step=int(value.step)
                    ))
                    average_value = average_values.get(value.step)
                    if average_value:
                        sec_scc.thermodynamics.append(Thermodynamics(
                            kinetic_energy=average_value.kinetic * n_atoms, potential_energy=average_value.potential * n_atoms,
                            time_step=int(average_value.step), pressure=average_value.pressure,
                            temperature=average_value.temperature
                        ))
                    trajectory = self.traj_parser.universe.trajectory[index]
                    if trajectory is not None and trajectory.has_forces:
                        sec_scc.forces = Forces(total=ForcesEntry(value=trajectory.forces * (ureg.kJ / ureg.angstrom)))

                    sec_scc.system_ref = sec_system
            # TODO add support for other tinker programs

            # control parameters
            # a key file can be specified optionally via the cli we assign this instead to get the parameters
            self.key_parser.mainfile = os.path.join(self.maindir, program.get('key', ''))
            if self.key_parser.mainfile is None:
                # if the key file is not specified in cli, read from either basename.key
                # or tinker.key
                self.key_parser.mainfile = self._get_tinker_file('key')

            parameters = {key.lower(): val for key, val in self.key_parser.get('key_val', [])}
            sec_run.x_tinker_control_parameters = parameters
            # TODO should this be removed and only have a dictionary of control parameters
            sec_control = sec_run.m_create(x_tinker_section_control_parameters)
            for key, val in parameters.items():
                key = key.replace('-', '_')
                setattr(sec_control, f'x_tinker_inout_control_{key}', val if isinstance(val, bool) else str(val))

            self.parse_method()

            self.parse_workflow(program, run)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import io
import shutil
from datetime import datetime
from typing import Tuple, List, Dict, Any, Optional
from pydantic import BaseModel, Field, validator
from mongoengine.queryset.visitor import Q
from fastapi import (
    APIRouter, Request, File, UploadFile, status, Depends, Path, Query as FastApiQuery,
    HTTPException)
from fastapi.responses import StreamingResponse
from fastapi.exceptions import RequestValidationError

from nomad import utils, config, files
from nomad.files import StagingUploadFiles, UploadBundle, is_safe_relative_path
from nomad.processing import Upload, Entry, ProcessAlreadyRunning, ProcessStatus, MetadataEditRequestHandler
from nomad.utils import strip
from nomad.search import search, refresh as search_refresh

from .auth import create_user_dependency, generate_upload_token
from ..models import (
    MetadataPagination, User, Direction, Pagination, PaginationResponse, HTTPExceptionModel,
    Files, files_parameters, WithQuery, MetadataEditRequest)
from .entries import EntryArchiveResponse, answer_entry_archive_request
from ..utils import (
    parameter_dependency_from_model, create_responses, DownloadItem, browser_download_headers,
    create_download_stream_zipped, create_download_stream_raw_file, create_stream_from_string)

router = APIRouter()
default_tag = 'uploads'
metadata_tag = 'uploads/metadata'
raw_tag = 'uploads/raw'
archive_tag = 'uploads/archive'
action_tag = 'uploads/action'
bundle_tag = 'uploads/bundle'

logger = utils.get_logger(__name__)


class ProcData(BaseModel):
    process_running: bool = Field(
        description='If a process is running')
    current_process: Optional[str] = Field(
        description='Name of the current or last completed process')
    process_status: str = Field(
        description='The status of the current or last completed process')
    last_status_message: Optional[str] = Field(
        description='A short, human readable message from the current process, with '
                    'information about what the current process is doing, or information '
                    'about the completion (successful or not) of the last process, if no '
                    'process is currently running.')
    errors: List[str] = Field(
        descriptions='A list of error messages that occurred during the last processing')
    warnings: List[str] = Field(
        description='A list of warning messages that occurred during the last processing')
    complete_time: Optional[datetime] = Field(
        description='Date and time of the completion of the last process')

    class Config:
        orm_mode = True


class UploadProcData(ProcData):
    upload_id: str = Field(
        None,
        description='The unique id for the upload.')
    upload_name: Optional[str] = Field(
        description='The name of the upload. This can be provided during upload '
                    'using the `upload_name` query parameter.')
    upload_create_time: datetime = Field(
        None,
        description='Date and time of the creation of the upload.')
    main_author: str = Field(
        None,
        description=strip('The main author of the upload.'))
    coauthors: List[str] = Field(
        None,
        description=strip('A list of upload coauthors.'))
    reviewers: List[str] = Field(
        None,
        description=strip('A user provided list of reviewers.'))
    viewers: List[str] = Field(
        None,
        description=strip('All viewers (main author, upload coauthors, and reviewers)'))
    writers: List[str] = Field(
        None,
        description=strip('All writers (main author, upload coauthors)'))
    published: bool = Field(
        False,
        description='If this upload is already published.')
    published_to: List[str] = Field(
        None,
        description='A list of other NOMAD deployments that this upload was uploaded to already.')
    publish_time: Optional[datetime] = Field(
        'Date and time of publication, if the upload has been published.')
    with_embargo: bool = Field(
        description='If the upload has an embargo set (embargo_length not equal to zero).')
    embargo_length: int = Field(
        description='The length of the requested embargo, in months. 0 if no embargo is requested.')
    license: str = Field(
        description='The license under which this upload is distributed.')
    entries: int = Field(
        0,
        description='The number of identified entries in this upload.')
    upload_files_server_path: Optional[str] = Field(
        None, description='The path to the uploads files on the server.')


class EntryProcData(ProcData):
    entry_id: str = Field()
    entry_create_time: datetime = Field()
    mainfile: str = Field()
    mainfile_key: Optional[str] = Field()
    upload_id: str = Field()
    parser_name: str = Field()
    entry_metadata: Optional[dict] = Field()


class UploadProcDataPagination(Pagination):
    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        if order_by is None:
            return 'upload_create_time'  # Default value
        assert order_by in ('upload_create_time', 'publish_time', 'upload_name', 'last_status_message'), 'order_by must be a valid attribute'
        return order_by

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        # Validation handled elsewhere
        return page_after_value


upload_proc_data_pagination_parameters = parameter_dependency_from_model(
    'upload_proc_data_pagination_parameters', UploadProcDataPagination)


class EntryProcDataPagination(Pagination):
    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        if order_by is None:
            return 'mainfile'  # Default value
        assert order_by in ('mainfile', 'parser_name', 'process_status', 'current_process'), 'order_by must be a valid attribute'
        return order_by

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        # Validation handled elsewhere
        return page_after_value


entry_proc_data_pagination_parameters = parameter_dependency_from_model(
    'entry_proc_data_pagination_parameters', EntryProcDataPagination)


class UploadProcDataResponse(BaseModel):
    upload_id: str = Field(None, description=strip('''
        Unique id of the upload.'''))
    data: UploadProcData = Field(
        None, description=strip('''
        The upload data as a dictionary.'''))


class UploadProcDataQuery(BaseModel):
    upload_id: Optional[List[str]] = Field(
        description='Search for uploads matching the given id. Multiple values can be specified.')
    upload_name: Optional[List[str]] = Field(
        description='Search for uploads matching the given upload_name. Multiple values can be specified.')
    is_processing: Optional[bool] = Field(
        description=strip('''
            If True, only include currently processing uploads.
            If False, do not include currently processing uploads.
            If unset, include everything.'''))
    is_published: Optional[bool] = Field(
        description=strip('''
            If True: only include published uploads.
            If False: only include unpublished uploads.
            If unset: include everything.'''))


upload_proc_data_query_parameters = parameter_dependency_from_model(
    'upload_proc_data_query_parameters', UploadProcDataQuery)


class UploadProcDataQueryResponse(BaseModel):
    query: UploadProcDataQuery = Field()
    pagination: PaginationResponse = Field()
    data: List[UploadProcData] = Field(
        None, description=strip('''
        The upload data as a list. Each item is a dictionary with the data for each
        upload.'''))


class EntryProcDataResponse(BaseModel):
    entry_id: str = Field()
    data: EntryProcData = Field()


class EntryProcDataQueryResponse(BaseModel):
    pagination: PaginationResponse = Field()
    processing_successful: int = Field(
        None, description=strip('''
        Number of entries that has been processed successfully.
        '''))
    processing_failed: int = Field(
        None, description=strip('''
        Number of entries that failed to process.
        '''))
    upload: UploadProcData = Field(
        None, description=strip('''
        The upload processing data of the upload.
        '''))
    data: List[EntryProcData] = Field(
        None, description=strip('''
        The entries data as a list. Each item is a dictionary with the data for one entry.
        '''))


class RawDirPagination(Pagination):
    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        assert not order_by, 'Cannot specify `order_by` for rawdir calls'
        return None

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        # Validation handled elsewhere
        return page_after_value


rawdir_pagination_parameters = parameter_dependency_from_model(
    'rawdir_pagination_parameters', RawDirPagination, exclude=['order', 'order_by'])


class RawDirFileMetadata(BaseModel):
    ''' Metadata about a file '''
    name: str = Field()
    size: Optional[int] = Field()
    entry_id: Optional[str] = Field(description=strip('''
        If this is a mainfile: the ID of the corresponding entry.'''))
    parser_name: Optional[str] = Field(description=strip('''
        If this is a mainfile: the name of the matched parser.'''))


class RawDirElementMetadata(RawDirFileMetadata):
    ''' Metadata about an directory *element*, i.e. a file or a directory '''
    is_file: bool = Field()


class RawDirDirectoryMetadata(BaseModel):
    ''' Metadata about a directory '''
    name: str = Field()
    size: Optional[int] = Field()
    content: List[RawDirElementMetadata] = Field(
        example=[
            {'name': 'a_directory', 'is_file': False, 'size': 456},
            {'name': 'a_file.json', 'is_file': True, 'size': 123, 'entry_id': 'XYZ', 'parser_name': 'parsers/vasp'}])


class RawDirResponse(BaseModel):
    path: str = Field(example='The/requested/path')
    access: str = Field()
    file_metadata: Optional[RawDirFileMetadata] = Field()
    directory_metadata: Optional[RawDirDirectoryMetadata] = Field()
    pagination: Optional[PaginationResponse] = Field()


class ProcessingData(BaseModel):
    upload_id: str = Field()
    path: str = Field()
    entry_id: Optional[str] = Field()
    parser_name: Optional[str] = Field()
    entry: Optional[EntryProcData] = Field()
    archive: Optional[Dict[str, Any]] = Field()


class PutRawFileResponse(BaseModel):
    upload_id: str = Field(None, description=strip('''
        Unique id of the upload.'''))
    data: UploadProcData = Field(None, description=strip('''
        The upload data as a dictionary.'''))
    processing: Optional[ProcessingData] = Field(None, description=strip('''
        Information about the processing, including the entry (if one was generated) and
        [optionally] the archive data of this entry.'''))


class UploadCommandExamplesResponse(BaseModel):
    upload_url: str = Field()
    upload_command: str = Field()
    upload_command_with_name: str = Field()
    upload_progress_command: str = Field()
    upload_command_form: str = Field()
    upload_tar_command: str = Field()


_not_authorized = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. Authorization is required, but no or bad authentication credentials provided.''')}

_not_authorized_to_upload = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. No credentials provided, or you do not have permissions to the
        specified upload.''')}

_not_authorized_to_entry = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. No credentials provided, or you do not have permissions to the
        specified upload or entry.''')}

_bad_request = status.HTTP_400_BAD_REQUEST, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Bad request. The request could not be processed because of some error/invalid argument.''')}

_bad_pagination = status.HTTP_400_BAD_REQUEST, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Bad request. Invalid pagination arguments supplied.''')}

_upload_not_found = status.HTTP_404_NOT_FOUND, {
    'model': HTTPExceptionModel,
    'description': strip('''
        The specified upload could not be found.''')}

_entry_not_found = status.HTTP_404_NOT_FOUND, {
    'model': HTTPExceptionModel,
    'description': strip('''
        The specified upload or entry could not be found.''')}

_upload_or_path_not_found = status.HTTP_404_NOT_FOUND, {
    'model': HTTPExceptionModel,
    'description': strip('''
        The specified upload, or a resource with the specified path within the upload,
        could not be found.''')}

_post_upload_response = 200, {
    'model': UploadProcDataResponse,
    'content': {
        'application/json': {},
        'text/plain': {'example': 'Thanks for uploading your data to nomad.'}
    },
    'description': strip('''
        A json structure with upload data, or a plain text information string.
        It will be a json structure if the request headers specifies `Accept = application/json`.''')}

_put_raw_file_response = 200, {
    'model': PutRawFileResponse,
    'content': {
        'application/json': {},
        'text/plain': {'example': 'Thanks for uploading your data to nomad.'}
    },
    'description': strip('''
        A json structure with upload data and possibly information from the processing,
        or a plain text information string.
        It will be a json structure if the request headers specifies `Accept = application/json`
        or if `wait_for_processing` is set.''')}


_raw_path_response = 200, {
    'content': {
        'application/octet-stream': {'example': 'file data'},
        'application/zip': {'example': '<zipped file or directory content>'}},
    'description': strip('''
        If `path` denotes a file: a stream with the file content, zipped if `compress = true`.
        If `path` denotes a directory, and `compress = true`, the directory content, zipped.''')}

_upload_bundle_response = 200, {
    'content': {
        'application/zip': {'example': '<zipped bundle data>'}}}


_no_name = 'NO NAME'

_thank_you_message = f'''
Thanks for uploading your data to nomad.
Go back to {config.gui_url()} and press
reload to see the progress on your upload
and publish your data.'''


@router.get(
    '/command-examples', tags=[default_tag],
    summary='Get example commands for shell based uploads.',
    response_model=UploadCommandExamplesResponse,
    responses=create_responses(_not_authorized),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_command_examples(user: User = Depends(create_user_dependency(required=True))):
    ''' Get url and example command for shell based uploads. '''
    token = generate_upload_token(user)
    api_url = config.api_url(ssl=config.services.https_upload, api='api/v1')
    upload_url = f'{api_url}/uploads?token={token}'
    upload_url_with_name = upload_url + '&upload_name=<name>'
    # Upload via streaming data tends to work much easier, e.g. no mime type issues, etc.
    # It is also easier for the user to unterstand IMHO.
    upload_command = f"curl -X POST '{upload_url}' -T <local_file>"
    rv = UploadCommandExamplesResponse(
        upload_url=upload_url,
        upload_command=upload_command,
        upload_command_form=f"curl -X POST '{upload_url}' -F file=@<local_file>",
        upload_command_with_name=f"curl -X POST '{upload_url_with_name}' -T <local_file>",
        upload_progress_command=upload_command + ' | xargs echo',
        upload_tar_command=f"tar -cf - <local_folder> | curl -# '{upload_url}' -X POST -T - | xargs echo"
    )
    return rv


@router.get(
    '', tags=[metadata_tag],
    summary='List uploads of authenticated user.',
    response_model=UploadProcDataQueryResponse,
    responses=create_responses(_not_authorized, _bad_pagination),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_uploads(
        request: Request,
        query: UploadProcDataQuery = Depends(upload_proc_data_query_parameters),
        pagination: UploadProcDataPagination = Depends(upload_proc_data_pagination_parameters),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Retrieves metadata about all uploads that match the given query criteria.
    '''
    # Build query
    mongo_query = Q()
    user_id = str(user.user_id)
    mongo_query &= Q(main_author=user_id) | Q(reviewers=user_id) | Q(coauthors=user_id)

    if query.upload_id:
        mongo_query &= Q(upload_id__in=query.upload_id)

    if query.upload_name:
        mongo_query &= Q(upload_name__in=query.upload_name)

    if query.is_processing is True:
        mongo_query &= Q(process_status__in=ProcessStatus.STATUSES_PROCESSING)
    elif query.is_processing is False:
        mongo_query &= Q(process_status__in=ProcessStatus.STATUSES_NOT_PROCESSING)

    if query.is_published is True:
        mongo_query &= Q(publish_time__ne=None)
    elif query.is_published is False:
        mongo_query &= Q(publish_time=None)

    # Fetch data from DB
    mongodb_query = Upload.objects.filter(mongo_query)
    # Create response
    start = pagination.get_simple_index()
    end = start + pagination.page_size

    order_by = pagination.order_by
    order_by_with_sign = order_by if pagination.order == Direction.asc else '-' + order_by
    if order_by == 'upload_create_time':
        order_by_args = [order_by_with_sign, 'upload_id']  # Use upload_id as tie breaker
    else:
        order_by_args = [order_by_with_sign, 'upload_create_time', 'upload_id']

    mongodb_query = mongodb_query.order_by(*order_by_args)

    data = [_upload_to_pydantic(upload) for upload in mongodb_query[start:end]]

    pagination_response = PaginationResponse(total=mongodb_query.count(), **pagination.dict())
    pagination_response.populate_simple_index_and_urls(request)

    return UploadProcDataQueryResponse(
        query=query,
        pagination=pagination_response,
        data=data)


@router.get(
    '/{upload_id}', tags=[metadata_tag],
    summary='Get a specific upload',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_upload(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload to retrieve.'),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Fetches a specific upload by its upload_id.
    '''
    # Get upload (or throw exception if nonexistent/no access)
    upload = _get_upload_with_read_access(upload_id, user)

    return UploadProcDataResponse(
        upload_id=upload_id,
        data=_upload_to_pydantic(upload))


@router.get(
    '/{upload_id}/entries', tags=[metadata_tag],
    summary='Get the entries of the specific upload as a list',
    response_model=EntryProcDataQueryResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_pagination),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_upload_entries(
        request: Request,
        upload_id: str = Path(
            ...,
            description='The unique id of the upload to retrieve entries for.'),
        pagination: EntryProcDataPagination = Depends(entry_proc_data_pagination_parameters),
        user: User = Depends(create_user_dependency())):
    '''
    Fetches the entries of a specific upload. Pagination is used to browse through the
    results.
    '''
    upload = _get_upload_with_read_access(upload_id, user, include_others=True)

    order_by = pagination.order_by
    order_by_with_sign = order_by if pagination.order == Direction.asc else '-' + order_by

    start = pagination.get_simple_index()
    end = start + pagination.page_size

    # load upload's entries. Use entry_id as tie breaker for ordering.
    entries = list(upload.entries_sublist(start, end, order_by=(order_by_with_sign, 'entry_id')))
    failed_entries_count = upload.failed_entries_count

    # load entries's metadata from search
    metadata_entries_query = WithQuery(
        query={
            'entry_id:any': list([entry.entry_id for entry in entries])
        }).query
    metadata_entries = search(
        pagination=MetadataPagination(page_size=len(entries)),
        owner='admin' if user and user.is_admin else 'visible',
        user_id=user.user_id if user else None,
        query=metadata_entries_query)
    metadata_entries_map = {
        metadata_entry['entry_id']: metadata_entry
        for metadata_entry in metadata_entries.data}

    # convert data to pydantic
    data = []
    for entry in entries:
        pydantic_entry = _entry_to_pydantic(entry)
        pydantic_entry.entry_metadata = metadata_entries_map.get(entry.entry_id)
        data.append(pydantic_entry)

    pagination_response = PaginationResponse(total=upload.total_entries_count, **pagination.dict())
    pagination_response.populate_simple_index_and_urls(request)

    return EntryProcDataQueryResponse(
        pagination=pagination_response,
        processing_successful=upload.processed_entries_count - failed_entries_count,
        processing_failed=failed_entries_count,
        upload=_upload_to_pydantic(upload),
        data=data)


@router.get(
    '/{upload_id}/entries/{entry_id}', tags=[metadata_tag],
    summary='Get a specific entry for a specific upload',
    response_model=EntryProcDataResponse,
    responses=create_responses(_entry_not_found, _not_authorized_to_entry),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_upload_entry(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload.'),
        entry_id: str = Path(
            ...,
            description='The unique id of the entry, belonging to the specified upload.'),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Fetches a specific entry for a specific upload.
    '''
    upload = _get_upload_with_read_access(upload_id, user)
    entry = upload.get_entry(entry_id)
    if not entry:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=strip('''
            An entry by that id could not be found in the specified upload.'''))

    data = _entry_to_pydantic(entry, add_es_metadata=True, user=user)

    return EntryProcDataResponse(entry_id=entry_id, data=data)


@router.get(
    '/{upload_id}/rawdir/{path:path}', tags=[raw_tag],
    summary='Get the metadata for the raw file or folder located at the specified path in the specified upload.',
    response_model=RawDirResponse,
    responses=create_responses(_upload_or_path_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_upload_rawdir_path(
        request: Request,
        upload_id: str = Path(
            ...,
            description='The unique id of the upload.'),
        path: str = Path(
            ...,
            description='The path within the upload raw files.'),
        pagination: RawDirPagination = Depends(rawdir_pagination_parameters),
        include_entry_info: bool = FastApiQuery(
            False,
            description=strip('''
                If the fields `entry_id` and `parser_name` should be populated for all
                encountered mainfiles.''')),
        user: User = Depends(create_user_dependency(required=False))):
    '''
    For the upload specified by `upload_id`, gets the raw file or directory metadata
    located at the given `path`. The response will either contain a `file_metadata` or
    `directory_metadata` key. For files, basic data about the file is returned, such as its
    name and size. For directories, the response includes a list of elements
    (files and folders) in the directory. For directories, the result is paginated.
    '''
    # Get upload
    upload = _get_upload_with_read_access(upload_id, user, include_others=True)
    upload_files = None
    try:
        # Get upload files
        upload_files = upload.upload_files
        if not upload_files.raw_path_exists(path):
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=strip('''
                Not found. Invalid path?'''))

        response = RawDirResponse(
            path=path.rstrip('/'),
            access='unpublished' if not upload.published else (
                'embargoed' if upload.embargo_length else 'public'))

        if upload_files.raw_path_is_file(path):
            # Path denotes a file
            response.file_metadata = RawDirFileMetadata(
                name=os.path.basename(path),
                size=upload_files.raw_file_size(path))
            if include_entry_info:
                entry: Entry = Entry.objects(upload_id=upload_id, mainfile=path, mainfile_key=None).first()
                if entry:
                    response.file_metadata.entry_id = entry.entry_id
                    response.file_metadata.parser_name = entry.parser_name
        else:
            # Path denotes a directory
            start = pagination.get_simple_index()
            end = start + pagination.page_size
            directory_list = upload_files.raw_directory_list(path)
            upload_files.close()
            content = []
            path_to_element: Dict[str, RawDirElementMetadata] = {}
            total = 0
            total_size = 0
            for i, path_info in enumerate(directory_list):
                total += 1
                total_size += path_info.size
                if start <= i < end:
                    element = RawDirElementMetadata(
                        name=os.path.basename(path_info.path),
                        is_file=path_info.is_file,
                        size=path_info.size)
                    content.append(element)
                    if include_entry_info:
                        path_to_element[path_info.path] = element

            if include_entry_info and content:
                for entry in Entry.objects(
                        upload_id=upload_id,
                        mainfile__in=path_to_element.keys(),
                        mainfile_key=None):
                    element = path_to_element[entry.mainfile]
                    element.entry_id = entry.entry_id
                    element.parser_name = entry.parser_name

            response.directory_metadata = RawDirDirectoryMetadata(
                name=os.path.basename(path),
                size=total_size,
                content=content)

            pagination_response = PaginationResponse(total=total, **pagination.dict())
            pagination_response.populate_simple_index_and_urls(request)
            response.pagination = pagination_response

        return response
    except Exception:
        if upload_files:
            upload_files.close()
        raise


@router.get(
    '/{upload_id}/raw/{path:path}', tags=[raw_tag],
    summary='Download the raw file or folder located at the specified path in the specified upload.',
    response_class=StreamingResponse,
    responses=create_responses(
        _raw_path_response, _upload_or_path_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_upload_raw_path(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload.'),
        path: str = Path(
            ...,
            description='The path within the upload raw files.'),
        files_params: Files = Depends(files_parameters),
        offset: Optional[int] = FastApiQuery(
            0,
            description=strip('''
                When dowloading individual files with `compress = false`, this can be
                used to seek to a specified position within the file in question. Default
                is 0, i.e. the start of the file.''')),
        length: Optional[int] = FastApiQuery(
            -1,
            description=strip('''
                When dowloading individual files with `compress = false`, this can be
                used to specify the number of bytes to read. By default, the value is -1,
                which means that the remainder of the file is streamed.''')),
        decompress: bool = FastApiQuery(
            False,
            description=strip('''
                Set if compressed files should be decompressed before streaming the
                content (that is: if there are compressed files *within* the raw files).
                Note, only some compression formats are supported.''')),
        ignore_mime_type: bool = FastApiQuery(
            False,
            description=strip('''
                Sets the mime type specified in the response headers to `application/octet-stream`
                instead of the actual mime type.''')),
        user: User = Depends(create_user_dependency(required=False, signature_token_auth_allowed=True))):
    '''
    For the upload specified by `upload_id`, gets the raw file or directory content located
    at the given `path`. The data is zipped if `compress = true`.

    It is possible to download both individual files and directories, but directories can
    only be downloaded if `compress = true`. When downloading a directory, it is also
    possible to specify `re_pattern`, `glob_pattern` or `include_files` to filter the files
    based on the file names.

    When downloading a file, you can specify `decompress` to attempt to decompress the data
    if the file is compressed before streaming it. You can also specify `offset` and `length`
    to download only a segment of the file (*Note:* `offset` and `length` does not work if
    `compress` is set to true).
    '''
    if files_params.compress and (offset != 0 or length != -1):
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
            Cannot specify `offset` or `length` when `compress` is true'''))
    # Get upload
    upload = _get_upload_with_read_access(upload_id, user, include_others=True)
    # Get upload files
    upload_files = upload.upload_files
    try:
        if not upload_files.raw_path_exists(path):
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=strip('''
                Not found. Invalid path?'''))
        if upload_files.raw_path_is_file(path):
            # File
            if files_params.compress:
                media_type = 'application/zip'
                download_item = DownloadItem(
                    upload_id=upload_id,
                    raw_path=path,
                    zip_path=os.path.basename(path))
                content = create_download_stream_zipped(
                    download_item, upload_files, compress=True)
            else:
                if offset < 0:
                    raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                        Invalid offset provided.'''))
                if length <= 0 and length != -1:
                    raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                        Invalid length provided. Should be greater than 0, or -1 if the remainder
                        of the file should be read.'''))
                if ignore_mime_type or not (offset == 0 and length == -1):
                    media_type = 'application/octet-stream'
                else:
                    media_type = upload_files.raw_file_mime_type(path)
                content = create_download_stream_raw_file(
                    upload_files, path, offset, length, decompress)
            return StreamingResponse(content, headers=browser_download_headers(
                filename=os.path.basename(path) + ('.zip' if files_params.compress else ''),
                media_type=media_type))
        else:
            # Directory
            if not files_params.compress:
                raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                    Path is a directory, `compress` must be set to true'''))
            # Stream directory content, compressed.
            download_item = DownloadItem(
                upload_id=upload_id, raw_path=path, zip_path='')
            content = create_download_stream_zipped(
                download_item, upload_files,
                re_pattern=files_params.re_pattern, recursive=True,
                create_manifest_file=False, compress=True)
            return StreamingResponse(content, headers=browser_download_headers(
                (upload.upload_id if not path else os.path.basename(path.rstrip('/'))) + '.zip',
                media_type='application/zip'))
    except Exception as e:
        logger.error('exception while streaming download', exc_info=e)
        upload_files.close()
        raise


@router.put(
    '/{upload_id}/raw/{path:path}', tags=[raw_tag],
    summary='Upload a raw file to the specified path (directory) in the specified upload.',
    response_class=StreamingResponse,
    responses=create_responses(
        _put_raw_file_response, _upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def put_upload_raw_path(
        request: Request,
        upload_id: str = Path(
            ...,
            description='The unique id of the upload.'),
        path: str = Path(
            ...,
            description='The path within the upload raw files.'),
        file: UploadFile = File(None),
        local_path: str = FastApiQuery(
            None,
            description=strip('''
            Internal/Admin use only.''')),
        file_name: str = FastApiQuery(
            None,
            description=strip('''
            Specifies the name of the file, when using method 2.''')),
        wait_for_processing: bool = FastApiQuery(
            False,
            description=strip('''
            Waits for the processing to complete and return information about the outcome
            in the response (**USE WITH CARE**).''')),
        include_archive: bool = FastApiQuery(
            False,
            description=strip('''
            If the archive data should be included in the response when using
            `wait_for_processing` (**USE WITH CARE**).''')),
        user: User = Depends(create_user_dependency(required=True, upload_token_auth_allowed=True))):
    '''
    Upload a file to the directory specified by `path` in the the upload specified by `upload_id`.

    When uploading a zip or tar archive, it will first be extracted, and the content will be
    *merged* with the existing content, i.e. new files are added, and if there is a collision
    (an old file with the same path and name as one of the new files), the old file will
    be overwritten, but the rest of the old files will remain untouched. If the file is not
    a zip or tar archive, the file will just be uploaded as it is, overwriting the existing
    file if there is one.

    The `path` should denote a directory. The empty string gives the "root" directory.

    If a single file is uploaded (i.e. not a zip or tar archive), it is possible to specify
    `wait_for_processing`. This means that the file (and only this file) will be matched and
    processed, and information about the outcome will be returned with the response. **NOTE**:
    this should be used with caution! When this option is set, the call will block until
    processing is complete, which may take some time. Also note, that just processing the
    new/modified file may not be enough in some cases (since adding/modifying a file somewhere
    in the directory structure may affect other entries). Also note that
    processing.entry.entry_metadata will not be populated in the response.

    There are two basic ways to upload a file: in the multipart-formdata or streaming the
    file data in the http body. Both are supported. Note, however, that the second method
    does not transfer a filename. If a transfer is made using method 2, you can specify
    the query argument `file_name` to name it. This *needs* to be specified when using
    method 2, unless you are uploading a zip/tar file (for zip/tar files the names don't
    matter since they are extracted). See the POST `uploads` endpoint for examples of curl
    commands for uploading files.
    '''
    if include_archive and not wait_for_processing:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='`include_archive` requires `wait_for_processing`.')

    upload = _get_upload_with_write_access(upload_id, user, include_published=False)

    if not is_safe_relative_path(path):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Bad path provided.')

    upload_path, method = await _get_file_if_provided(
        upload_id, request, file, local_path, file_name, user)

    if not upload_path:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='No upload file provided.')

    is_compressed = files.zipfile.is_zipfile(upload_path) or files.tarfile.is_tarfile(upload_path)

    if not wait_for_processing:
        # Process on worker (normal case)
        if is_compressed:
            # Uploading an compressed file -> reprocess the entire target directory
            path_filter = path
        else:
            # Uploading a single file -> reprocess only the file
            path_filter = os.path.join(path, os.path.basename(upload_path))

        try:
            upload.process_upload(
                file_operation=dict(op='ADD', path=upload_path, target_dir=path, temporary=(method != 0)),
                path_filter=path_filter)
        except ProcessAlreadyRunning:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='The upload is currently blocked by another process.')

        if request.headers.get('Accept') == 'application/json':
            response = PutRawFileResponse(
                upload_id=upload_id,
                data=_upload_to_pydantic(upload))
            response_text = response.json()
            media_type = 'application/json'
        else:
            response_text = _thank_you_message
            media_type = 'text/plain'
    else:
        # Process locally
        if is_compressed:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='`wait_for_processing` can only be used with single files, not with compressed files.')

        full_path = os.path.join(path, os.path.basename(upload_path))
        try:
            reprocess_settings = dict(
                index_invidiual_entries=True, reprocess_existing_entries=True)
            entry = upload.put_file_and_process_local(
                upload_path, path, reprocess_settings=reprocess_settings)
            if upload.process_status == ProcessStatus.FAILURE:
                # Should only happen if we fail to put the file, match the file, or to *initiate*
                # entry processing - i.e. normally, this shouldn't happen, not even with
                # an badly formatted/unparsable mainfile.
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=f'Failed to put and process: {upload.errors[0]}')

            search_refresh()

            archive = None
            if entry and entry.process_status == ProcessStatus.SUCCESS and include_archive:
                # NOTE: We can't rely on ES to get the metadata for the entry, since it may
                # not have hade enough time to update its index etc. For now, we will just
                # ignore this, as we do not need it.
                entry_metadata = dict(
                    upload_id=upload_id,
                    entry_id=entry.entry_id,
                    parser_name=entry.parser_name)
                archive = answer_entry_archive_request(
                    dict(upload_id=upload_id, mainfile=full_path),
                    required='*', user=user,
                    entry_metadata=entry_metadata)['data']['archive']

            response = PutRawFileResponse(
                upload_id=upload_id,
                data=_upload_to_pydantic(upload),
                processing=ProcessingData(
                    upload_id=upload_id,
                    path=full_path,
                    entry_id=entry.entry_id if entry else None,
                    parser_name=entry.parser_name if entry else None,
                    entry=_entry_to_pydantic(entry) if entry else None,
                    archive=archive))
            response_text = response.json()
            media_type = 'application/json'

        except HTTPException:
            raise
        except ProcessAlreadyRunning:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='The upload is currently being processed, operation not allowed.')
        except Exception as e:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=f'Unexpected exception occurred: {e}')
        finally:
            try:
                shutil.rmtree(os.path.dirname(upload_path))
            except Exception:
                pass

    return StreamingResponse(create_stream_from_string(response_text), media_type=media_type)


@router.delete(
    '/{upload_id}/raw/{path:path}', tags=[raw_tag],
    summary='Delete the raw file or folder located at the specified path in the specified upload.',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def delete_upload_raw_path(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload.'),
        path: str = Path(
            ...,
            description='The path within the upload raw files.'),
        user: User = Depends(create_user_dependency(required=True, upload_token_auth_allowed=True))):
    '''
    Delete file or folder located at the specified path in the specified upload. The upload
    must not be published. This also automatically triggers a reprocessing of the upload.
    Choosing the empty string as `path` deletes all files.
    '''
    upload = _get_upload_with_write_access(upload_id, user, include_published=False)

    if not is_safe_relative_path(path):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Bad path provided.')

    upload_files = StagingUploadFiles(upload_id)

    if not upload_files.raw_path_exists(path):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail='No file or folder with that path found.')

    try:
        upload.process_upload(file_operation=dict(op='DELETE', path=path), path_filter=path)
    except ProcessAlreadyRunning:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The upload is currently blocked by another process.')

    return UploadProcDataResponse(upload_id=upload_id, data=_upload_to_pydantic(upload))


@router.post(
    '/{upload_id}/raw-create-dir/{path:path}', tags=[raw_tag],
    summary='Create a new empty directory with the specified path in the specified upload.',
    response_model=UploadProcDataResponse,
    responses=create_responses(
        _upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload_raw_create_dir_path(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload.'),
        path: str = Path(
            ...,
            description='The path within the upload raw files.'),
        user: User = Depends(create_user_dependency(required=True, upload_token_auth_allowed=True))):
    '''
    Create a new empty directory in the specified upload. The `path` should be the full path
    to the new directory (i.e. ending with the name of the new directory). The api call returns
    immediately (no processing is neccessary).
    '''
    upload = _get_upload_with_write_access(upload_id, user, include_published=False)

    if not path or not is_safe_relative_path(path):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Bad path provided.')
    if upload.staging_upload_files.raw_path_exists(path):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f'Path `{path}` already exists.')
    try:
        upload.staging_upload_files.raw_create_directory(path)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f'Failed to create directory: {e}')

    return UploadProcDataResponse(upload_id=upload_id, data=_upload_to_pydantic(upload))


@router.get(
    '/{upload_id}/archive/mainfile/{mainfile:path}', tags=[archive_tag],
    summary='Get the full archive for the given upload and mainfile path.',
    response_model=EntryArchiveResponse,
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    responses=create_responses(_upload_or_path_not_found, _not_authorized_to_upload))
async def get_upload_entry_archive_mainfile(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload.'),
        mainfile: str = Path(
            ...,
            description='The mainfile path within the upload\'s raw files.'),
        mainfile_key: Optional[str] = FastApiQuery(
            None,
            description='The mainfile_key, for accessing child entries.'),
        user: User = Depends(create_user_dependency(required=False))):
    '''
    For the upload specified by `upload_id`, gets the full archive of a single entry that
    is identified by the given `mainfile`.
    '''
    _get_upload_with_read_access(upload_id, user, include_others=True)
    query = dict(upload_id=upload_id, mainfile=mainfile)
    if mainfile_key:
        query.update(mainfile_key=mainfile_key)
    return answer_entry_archive_request(query, required='*', user=user)


@router.get(
    '/{upload_id}/archive/{entry_id}', tags=[archive_tag],
    summary='Get the full archive for the given upload and entry.',
    response_model=EntryArchiveResponse,
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    responses=create_responses(_upload_or_path_not_found, _not_authorized_to_upload))
async def get_upload_entry_archive(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload.'),
        entry_id: str = Path(
            ...,
            description='The unique entry id.'),
        user: User = Depends(create_user_dependency(required=False))):
    '''
    For the upload specified by `upload_id`, gets the full archive of a single entry that
    is identified by the given `entry_id`.
    '''
    _get_upload_with_read_access(upload_id, user, include_others=True)
    return answer_entry_archive_request(
        dict(upload_id=upload_id, entry_id=entry_id),
        required='*', user=user)


@router.post(
    '', tags=[default_tag],
    summary='Submit a new upload',
    response_class=StreamingResponse,
    responses=create_responses(_post_upload_response, _not_authorized, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload(
        request: Request,
        file: UploadFile = File(None),
        local_path: str = FastApiQuery(
            None,
            description=strip('''
            Internal/Admin use only.''')),
        file_name: str = FastApiQuery(
            None,
            description=strip('''
            Specifies the name of the file, when using method 2.''')),
        upload_name: str = FastApiQuery(
            None,
            description=strip('''
            A human readable name for the upload.''')),
        embargo_length: int = FastApiQuery(
            0,
            description=strip('''
            The requested embargo length, in months, if any (0-36).''')),
        publish_directly: bool = FastApiQuery(
            None,
            description=strip('''
            If the upload should be published directly. False by default.''')),
        user: User = Depends(create_user_dependency(required=True, upload_token_auth_allowed=True))):
    '''
    Creates a new, empty upload and, optionally, uploads a first file to it. If a file is
    provided, and it is a zip or tar file, it will first be extracted, then added.

    It is recommended to give the upload itself a descriptive `upload_name`. If not specified,
    it will be set to the file name (if provided). The `upload_name` can be edited afterwards (as
    long as the upload is not published).

    There are two basic ways to upload a file: in the multipart-formdata or streaming the
    file data in the http body. Both are supported. Note, however, that the second method
    does not transfer a filename. If a transfer is made using method 2, you can specify
    the query argument `file_name` to name it. This *needs* to be specified when using
    method 2, unless you are uploading a zip/tar file (for zip/tar files the names don't
    matter since they are extracted).

    Example curl commands for creating an upload and uploading a file:

    Method 1: multipart-formdata

        curl -X 'POST' "url" -F file=@local_file

    Method 2: streaming data

        curl -X 'POST' "url" -T local_file

    Authentication is required. This can either be done using the regular bearer token,
    or using the simplified upload token. To use the simplified upload token, just
    specify it as a query parameter in the url, i.e.

        curl -X 'POST' "baseurl?token=ABC.XYZ" ...

    Note, there is a limit on how many unpublished uploads a user can have. If exceeded,
    error code 400 will be returned.
    '''
    if not user.is_admin:
        # Check upload limit
        if _query_mongodb(main_author=str(user.user_id), publish_time=None).count() >= config.services.upload_limit:  # type: ignore
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                Limit of unpublished uploads exceeded for user.'''))

    if not 0 <= embargo_length <= 36:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='`embargo_length` must be between 0 and 36 months.')

    upload_id = utils.create_uuid()

    if upload_name and not file_name and files.is_safe_basename(upload_name):
        # Try to default the file_name using name
        file_name = upload_name

    upload_path, method = await _get_file_if_provided(
        upload_id, request, file, local_path, file_name, user)

    if not upload_name:
        # Try to default upload_name
        if method == 2:
            upload_name = file_name or None
        elif upload_path:
            upload_name = os.path.basename(upload_path)

    upload = Upload.create(
        upload_id=upload_id,
        main_author=user,
        upload_name=upload_name,
        upload_create_time=datetime.utcnow(),
        embargo_length=embargo_length,
        publish_directly=publish_directly)

    # Create staging files
    files.StagingUploadFiles(upload_id=upload_id, create=True)

    logger.info('upload created', upload_id=upload_id)

    if upload_path:
        upload.process_upload(
            file_operation=dict(op='ADD', path=upload_path, target_dir='', temporary=(method != 0)))

    if request.headers.get('Accept') == 'application/json':
        upload_proc_data_response = UploadProcDataResponse(
            upload_id=upload_id,
            data=_upload_to_pydantic(upload))
        response_text = upload_proc_data_response.json()
        media_type = 'application/json'
    else:
        response_text = _thank_you_message
        media_type = 'text/plain'

    return StreamingResponse(create_stream_from_string(response_text), media_type=media_type)


@router.post(
    '/{upload_id}/edit', tags=[metadata_tag],
    summary='Updates the metadata of the specified upload.',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload_edit(
        request: Request,
        data: MetadataEditRequest,
        upload_id: str = Path(..., description='The unique id of the upload.'),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Updates the metadata of the specified upload and entries. An optional `query` can be
    specified to select only some of the entries of the upload (the query results are
    automatically restricted to the specified upload).

    **Note:**
      - Only admins can edit some of the fields.
      - The embargo of a published upload is lifted by setting the `embargo_length` attribute
        to 0.
      - If the upload is published, the only operations permitted using this endpoint is to
        lift the embargo, i.e. set `embargo_length` to 0, and to edit the entries in datasets
        that where created by the current user.
      - If a query is specified, it is not possible to edit upload level metadata (like
        `upload_name`, `coauthors`, etc.), as the purpose of queries is to select only a
        subset of the upload entries to edit, but changing upload level metadata would affect
        **all** entries of the upload.
    '''
    edit_request_json = await request.json()
    try:
        MetadataEditRequestHandler.edit_metadata(edit_request_json, upload_id, user)
        return UploadProcDataResponse(upload_id=upload_id, data=_upload_to_pydantic(Upload.get(upload_id)))
    except RequestValidationError as e:
        raise  # A problem which we have handled explicitly. Fastapi does json conversion.
    except Exception as e:
        # The upload is processing or some kind of unexpected error has occured
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))


@router.delete(
    '/{upload_id}', tags=[default_tag],
    summary='Delete an upload',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def delete_upload(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload to delete.'),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Delete an existing upload.

    Only uploads that are sill in staging, not already deleted, not still uploaded, and
    not currently processed, can be deleted.
    '''
    upload = _get_upload_with_write_access(
        upload_id, user, include_published=True, published_requires_admin=True)

    try:
        upload.delete_upload()
    except ProcessAlreadyRunning:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
            The upload is still being processed.'''))
    except Exception as e:
        logger.error('could not delete processing upload', exc_info=e)
        raise

    return UploadProcDataResponse(
        upload_id=upload_id,
        data=_upload_to_pydantic(upload))


@router.post(
    '/{upload_id}/action/publish', tags=[action_tag],
    summary='Publish an upload',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload_action_publish(
        upload_id: str = Path(
            ...,
            description=strip('''
                The unique id of the upload to publish.''')),
        embargo_length: int = FastApiQuery(
            None,
            description=strip('''
                If provided, updates the embargo length of the upload. The value should
                be between 0 and 36 months. 0 means no embargo.''')),
        to_central_nomad: bool = FastApiQuery(
            False,
            description=strip('''
                Will send the upload to the central NOMAD repository and publish it. This
                option is only available on an OASIS. The upload must already be published
                on the OASIS.''')),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Publishes an upload. The upload cannot be modified after this point (except for special
    cases, like when lifting the embargo prematurely, and by admins). After the upload is
    published and the embargo period (if any) is expired, the generated archive entries
    will be publicly visible.
    '''
    upload = _get_upload_with_write_access(
        upload_id, user, include_published=True, published_requires_admin=False)

    if upload.published and not user.is_admin and not to_central_nomad:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Upload already published.')

    _check_upload_not_processing(upload)

    if upload.process_status == ProcessStatus.FAILURE:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Cannot publish an upload that failed processing.')
    if upload.processed_entries_count == 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Cannot publish an upload without any resulting entries.')
    if embargo_length is not None and not 0 <= embargo_length <= 36:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Invalid embargo_length. Must be between 0 and 36 months.')

    if to_central_nomad:
        # Publish from an OASIS to the central repository
        if not config.keycloak.oasis:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='Must be on an OASIS to publish to the central NOMAD repository.')
        if not upload.published:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='The upload must be published on the OASIS first.')
        # Everything looks ok, try to publish it to the central NOMAD!
        upload.publish_externally(embargo_length=embargo_length)
    else:
        # Publish to this repository
        if upload.published:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='The upload is already published.')
        try:
            upload.publish_upload(embargo_length=embargo_length)
        except ProcessAlreadyRunning:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='The upload is still/already processed.')

    return UploadProcDataResponse(
        upload_id=upload_id,
        data=_upload_to_pydantic(upload))


@router.post(
    '/{upload_id}/action/process', tags=[action_tag],
    summary='Manually triggers processing of an upload.',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload_action_process(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload to process.'),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Processes an upload, i.e. parses the files and updates the NOMAD archive. Only admins
    can process an already published upload.
    '''
    upload = _get_upload_with_write_access(
        upload_id, user, include_published=True, published_requires_admin=True)

    _check_upload_not_processing(upload)

    upload.process_upload()
    return UploadProcDataResponse(
        upload_id=upload_id,
        data=_upload_to_pydantic(upload))


@router.post(
    '/{upload_id}/action/lift-embargo', tags=[action_tag],
    summary='Lifts the embargo of an upload.',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload_action_lift_embargo(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload to lift the embargo for.'),
        user: User = Depends(create_user_dependency(required=True))):
    ''' Lifts the embargo of an upload. '''
    upload = _get_upload_with_write_access(
        upload_id, user, include_published=True, published_requires_admin=False)
    _check_upload_not_processing(upload)
    if not upload.published:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
            Upload is not published, no embargo to lift.'''))
    if not upload.with_embargo:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
            Upload has no embargo.'''))
    # Lift the embargo using MetadataEditRequestHandler.edit_metadata
    try:
        MetadataEditRequestHandler.edit_metadata({'metadata': {'embargo_length': 0}}, upload_id, user)
        upload.reload()
        return UploadProcDataResponse(
            upload_id=upload_id,
            data=_upload_to_pydantic(upload))
    except Exception as e:
        # Should only happen if the upload just started processing or something unexpected happens
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))


@router.get(
    '/{upload_id}/bundle', tags=[bundle_tag],
    summary='Gets an *upload bundle* for the specified upload.',
    response_class=StreamingResponse,
    responses=create_responses(
        _upload_bundle_response, _upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_upload_bundle(
        upload_id: str = Path(
        ...,
        description='The unique id of the upload.'),
        include_raw_files: Optional[bool] = FastApiQuery(
            True,
            description=strip('''
                If raw files should be included in the bundle (true by default).''')),
        include_archive_files: Optional[bool] = FastApiQuery(
            True,
            description=strip('''
                If archive files (i.e. parsed entries data) should be included in the bundle
                (true by default).''')),
        include_datasets: Optional[bool] = FastApiQuery(
            True,
            description=strip('''
                If datasets references to this upload should be included in the bundle
                (true by default).''')),
        user: User = Depends(create_user_dependency(required=False))):
    '''
    Get an *upload bundle* for the specified upload. An upload bundle is a file bundle which
    can be used to export and import uploads between different NOMAD deployments.
    '''
    upload = _get_upload_with_read_access(upload_id, user, include_others=True)
    _check_upload_not_processing(upload)

    try:
        stream = upload.export_bundle(
            export_as_stream=True, export_path=None, zipped=True, move_files=False, overwrite=False,
            include_raw_files=include_raw_files, include_archive_files=include_archive_files,
            include_datasets=include_datasets)
    except Exception as e:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
            Could not export due to error: ''' + str(e)))

    return StreamingResponse(stream, media_type='application/zip')


@router.post(
    '/bundle', tags=[bundle_tag],
    summary='Posts an *upload bundle* to this NOMAD deployment.',
    response_model=UploadProcDataResponse,
    responses=create_responses(_not_authorized, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload_bundle(
        request: Request,
        file: UploadFile = File(None),
        local_path: str = FastApiQuery(
            None,
            description=strip('''
            Internal/Admin use only.''')),
        embargo_length: Optional[int] = FastApiQuery(
            None,
            description=strip('''
                Specifies the embargo length in months to set on the upload. If omitted,
                the value specified in the bundle will be used. A value of 0 means no
                embargo.''')),
        include_raw_files: Optional[bool] = FastApiQuery(
            None,
            description=strip('''
                If raw files should be imported from the bundle
                *(only admins can change this setting)*.''')),
        include_archive_files: Optional[bool] = FastApiQuery(
            None,
            description=strip('''
                If archive files (i.e. parsed entries data) should be imported from the bundle
                *(only admins can change this setting)*.''')),
        include_datasets: Optional[bool] = FastApiQuery(
            None,
            description=strip('''
                If dataset references to this upload should be imported from the bundle
                *(only admins can change this setting)*.''')),
        include_bundle_info: Optional[bool] = FastApiQuery(
            None,
            description=strip('''
                If the bundle_info.json file should be kept
                *(only admins can change this setting)*.''')),
        keep_original_timestamps: Optional[bool] = FastApiQuery(
            None,
            description=strip('''
                If all original timestamps, including `upload_create_time`, `entry_create_time`
                and `publish_time`, should be kept
                *(only admins can change this setting)*.''')),
        set_from_oasis: Optional[bool] = FastApiQuery(
            None,
            description=strip('''
                If the `from_oasis` flag and `oasis_deployment_id` should be set
                *(only admins can change this setting)*.''')),
        trigger_processing: Optional[bool] = FastApiQuery(
            None,
            description=strip('''
                If processing should be triggered after the bundle has been imported
                *(only admins can change this setting)*.''')),
        user: User = Depends(create_user_dependency(required=True, upload_token_auth_allowed=True))):
    '''
    Posts an *upload bundle* to this NOMAD deployment. An upload bundle is a file bundle which
    can be used to export and import uploads between different NOMAD installations. The
    endpoint expects an upload bundle attached as a zipfile.

    **NOTE:** This endpoint is restricted to admin users and oasis admins. Further, all
    settings except `embargo_length` requires an admin user to change (these settings
    have default values specified by the system configuration).

    There are two basic ways to upload a file: in the multipart-formdata or streaming the
    file data in the http body. Both are supported. See the POST `uploads` endpoint for
    examples of curl commands for uploading files.
    '''
    is_admin = user.is_admin
    is_oasis = not is_admin and user.is_oasis_admin and config.bundle_import.allow_bundles_from_oasis

    if not is_admin and not is_oasis:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail='User not authorized to upload bundles')

    bundle_path, method = await _get_file_if_provided(
        tmp_dir_prefix='bundle', request=request, file=file, local_path=local_path, file_name=None, user=user)

    if not bundle_path:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail='No bundle file provided')

    try:
        bundle: UploadBundle = None
        bundle = UploadBundle(bundle_path)

        if is_oasis and not config.bundle_import.allow_unpublished_bundles_from_oasis:
            bundle_info = bundle.bundle_info
            if not bundle_info.get('upload', {}).get('publish_time'):
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=f'Bundles uploaded from an oasis must be published in the oasis first.')

        settings_dict: Dict[str, Any] = dict(
            include_raw_files=include_raw_files,
            include_archive_files=include_archive_files,
            include_datasets=include_datasets,
            include_bundle_info=include_bundle_info,
            keep_original_timestamps=keep_original_timestamps,
            set_from_oasis=set_from_oasis,
            trigger_processing=trigger_processing)

        for k, v in settings_dict.copy().items():
            if v is None:
                del settings_dict[k]
            elif v is not None and not is_admin:
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=f'Specifying setting {k} requires an admin user')

        upload = Upload.create_skeleton_from_bundle(bundle)
        bundle.close()
        upload.import_bundle(
            bundle_path, move_files=False, embargo_length=embargo_length,
            settings=settings_dict)

        return UploadProcDataResponse(
            upload_id=upload.upload_id,
            data=_upload_to_pydantic(upload))

    except Exception as e:
        if bundle:
            bundle.close()
        if method != 0:
            bundle.delete(include_parent_folder=True)
        if isinstance(e, HTTPException):
            raise
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail='Failed to import bundle: ' + str(e))


async def _get_file_if_provided(
        tmp_dir_prefix: str, request: Request, file: UploadFile, local_path: str, file_name: str,
        user: User) -> Tuple[str, int]:
    '''
    If the user provides a file with the api call, load it and save it to a temporary
    folder (or, if method 0 is used, forward the file). The method thus needs to identify
    which file transfer method was used (0 - 2), and save the data (if method is 1 or 2).

    Returns the os path to the resulting file and method (0-2), or (None, None) if no file
    data was provided with the api call.
    '''
    # Determine the source data stream
    file_name_argument_not_given = not file_name
    src_stream = None
    if local_path:
        # Method 0: Local file - only for admins
        if not user.is_admin:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
                You need to be admin to use local_path as method of upload.'''))
        if not os.path.exists(local_path) or not os.path.isfile(local_path):
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                The specified local_path cannot be found or is not a file.'''))
        method = 0
        file_name = os.path.basename(local_path)
    elif file:
        # Method 1: Data provided as formdata
        method = 1
        file_name = file.filename or _no_name
        src_stream = _asyncronous_file_reader(file)
    else:
        # Method 2: Data has to be sent streamed in the body
        method = 2
        file_name = file_name or _no_name
        src_stream = request.stream()

    if not files.is_safe_basename(file_name):
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
            Bad file name provided.'''))

    # Read the stream and save to file
    if method == 0:
        upload_path = local_path  # Use the provided path
        uploaded_bytes = os.path.getsize(local_path)
    else:
        tmp_dir = files.create_tmp_dir(tmp_dir_prefix)
        upload_path = os.path.join(tmp_dir, file_name)
        try:
            with open(upload_path, 'wb') as f:
                uploaded_bytes = 0
                log_interval = 1e9
                log_unit = 'GB'
                next_log_at = log_interval
                async for chunk in src_stream:
                    if not chunk:
                        # End of data stream
                        break
                    uploaded_bytes += len(chunk)
                    f.write(chunk)
                    if uploaded_bytes > next_log_at:
                        logger.info('Large upload in progress - uploaded: '
                                    f'{uploaded_bytes // log_interval} {log_unit}')
                        next_log_at += log_interval
                logger.info(f'Uploaded {uploaded_bytes} bytes')
        except Exception as e:
            if not (isinstance(e, RuntimeError) and 'Stream consumed' in str(e)):
                if os.path.exists(tmp_dir):
                    shutil.rmtree(tmp_dir)
                logger.warn('IO error receiving upload data', exc_info=e)
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail='Some IO went wrong, upload probably aborted/disrupted.')

    if not uploaded_bytes and method == 2:
        # No data was provided
        shutil.rmtree(tmp_dir)
        return None, None

    logger.info('received uploaded file')
    if method == 2 and file_name_argument_not_given:
        if not files.zipfile.is_zipfile(upload_path) and not files.tarfile.is_tarfile(upload_path):
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                Using method 2 and file does not look like a zip or tar file - must specify `file_name`.'''))

    return upload_path, method


async def _asyncronous_file_reader(f):
    ''' Asynchronous generator to read file-like objects. '''
    while True:
        try:
            data: bytes = await f.read(io.DEFAULT_BUFFER_SIZE)
        except Exception:
            await f.close()
            raise
        if not data:
            await f.close()
            return
        yield data


def _query_mongodb(**kwargs):
    return Upload.objects(**kwargs)


def _get_upload_with_read_access(upload_id: str, user: User, include_others: bool = False) -> Upload:
    '''
    Determines if the specified user has read access to the specified upload. If so, the
    corresponding Upload object is returned. If the upload does not exist, or the user has
    no read access to it, a HTTPException is raised.

    Arguments:
        upload_id: The id of the requested upload.
        user: The authenticated user, if any.
        include_others: If uploads owned by others should be included. Access to the uploads
            of other users is only granted if the upload is published and not under embargo.
    '''
    mongodb_query = _query_mongodb(upload_id=upload_id)
    upload = mongodb_query.first()
    if upload is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=strip('''
            The specified upload_id was not found.'''))
    if user and (user.is_admin or (str(user.user_id) in upload.viewers)):
        # Ok, the user a viewer, or we have an admin user
        return upload
    elif include_others:
        if not upload.published:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
                You do not have access to the specified upload - not published yet.'''))
        if upload.published and upload.with_embargo:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
                You do not have access to the specified upload - published with embargo.'''))
        return upload
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            You do not have access to the specified upload.'''))


def _get_upload_with_write_access(
        upload_id: str, user: User, include_published: bool = False,
        published_requires_admin: bool = True) -> Upload:
    '''
    Determines if the specified user has write access to the specified upload. If so, the
    corresponding Upload object is returned. If the upload does not exist, or the user has
    no write access to it, a HTTPException is raised.
    '''
    if not user:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            User authentication required to access uploads.'''))
    mongodb_query = _query_mongodb(upload_id=upload_id)
    if not mongodb_query.count():
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=strip('''
            The specified upload_id was not found.'''))
    upload = mongodb_query.first()
    if not user.is_admin and upload.main_author != str(user.user_id):
        if not upload.coauthors or str(user.user_id) not in upload.coauthors:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
                You do not have write access to the specified upload.'''))
    if upload.published:
        if not include_published:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
                Upload is already published, operation not possible.'''))
        if published_requires_admin and not user.is_admin:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
                Upload is already published, only admins can perform this operation.'''))
    return upload


def _upload_to_pydantic(upload: Upload) -> UploadProcData:
    ''' Converts the mongo db object to an UploadProcData object. '''
    pydantic_upload = UploadProcData.from_orm(upload)
    pydantic_upload.entries = upload.total_entries_count
    try:
        pydantic_upload.upload_files_server_path = upload.upload_files.external_os_path
    except KeyError:
        # In case the files are missing for one reason or another
        pass

    return pydantic_upload


def _entry_to_pydantic(entry: Entry, add_es_metadata: bool = False, user=None) -> EntryProcData:
    '''
    Converts the mongo db object to an EntryProcData object, and optionally also adds metadata
    from ES
    '''
    rv = EntryProcData.from_orm(entry)
    if add_es_metadata:
        # load entries's metadata from search
        metadata_entries = search(
            pagination=MetadataPagination(page_size=1),
            owner='admin' if user.is_admin else 'visible',
            user_id=user.user_id,
            query=dict(entry_id=entry.entry_id))
        if len(metadata_entries.data) == 1:
            rv.entry_metadata = metadata_entries.data[0]
    return rv


def _check_upload_not_processing(upload: Upload):
    '''
    Checks if the upload is processing, and raises a HTTPException (err code 400) if so.
    '''
    if upload.process_running:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The upload is currently being processed, operation not allowed.')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
A command that runs some example operations on a working nomad@FAIRDI installation
as a final integration test.
'''

import time
import os
import json

from nomad.client import api


def integrationtests(auth: api.Auth, skip_parsers: bool, skip_publish: bool, skip_doi: bool):
    multi_code_example_file = 'tests/data/integration/multi_code_data.zip'
    simple_example_file = 'tests/data/integration/examples_vasp.zip'
    has_doi = False
    published = False

    print('get the upload command')
    response = api.get('uploads/command-examples', auth=auth)
    assert response.status_code == 200, response.text
    command = response.json()['upload_command_with_name']

    def get_upload(upload):
        first = True
        while first or upload['process_running']:
            first = False
            response = api.get(f'uploads/{upload["upload_id"]}', auth=auth)
            if response.status_code == 404:
                return None
            assert response.status_code == 200, response.text
            upload = response.json()['data']
            time.sleep(0.3)

        return upload

    response = api.get('uploads', params=dict(upload_name='integration_test_upload'), auth=auth)
    assert response.status_code == 200, response.text
    uploads = response.json()['data']
    assert len(uploads) == 0, 'the test upload must not exist before'

    if not skip_parsers:
        print('upload multi code test data with curl')
        command = command.replace('<local_file>', multi_code_example_file)
        command = command.replace('<name>', 'integration_test_upload')
        command += ' -k'
        code = os.system(command)
        assert code == 0, 'curl command must be successful'
        response = api.get('uploads', params=dict(upload_name='integration_test_upload'), auth=auth)
        assert response.status_code == 200, response.text
        response_json = response.json()
        assert len(response_json['data']) == 1, 'exactly one test upload must be on the server'
        upload = response_json['data'][0]

        print('observe the upload process to be finished')
        upload = get_upload(upload)

        assert upload['process_status'] == 'SUCCESS'

        print('delete the upload again')
        upload = api.delete(f'uploads/{upload["upload_id"]}', auth=auth).json()['data']
        upload = get_upload(upload)

    print('upload simple data with API')
    with open(simple_example_file, 'rb') as f:
        response = api.post(
            'uploads', files=dict(file=f), params=dict(upload_name='integration_test_upload'),
            auth=auth, headers={'Accept': 'application/json'})
        assert response.status_code == 200, response.text
        upload = response.json()['data']

    print('observe the upload process to be finished')
    upload = get_upload(upload)
    response = api.get(f'uploads/{upload["upload_id"]}/entries', auth=auth)
    assert response.status_code == 200, response.text
    entries = response.json()['data']
    assert upload['entries'] == len(entries)

    try:
        print('get repo data')
        for entry in entries:
            response = api.get(f'entries/{entry["entry_id"]}', auth=auth)
            assert response.status_code == 200, response.text
            entry_metadata = response.json()['data']
            entry_metadata['entry_id'] == entry['entry_id']

        print('get archive data')
        for entry in entries:
            api.get(f'entries/{entry["entry_id"]}/archive/download', auth=auth)
            assert response.status_code == 200, response.text

        print('get archive logs')
        for entry in entries:
            response = api.post(
                f'entries/{entry["entry_id"]}/archive/query',
                data=json.dumps({
                    'required': {
                        'processing_logs': '*'
                    }
                }), auth=auth)
            assert response.status_code == 200, response.text
            assert list(response.json()['data']['archive'].keys()) == ['processing_logs']

        query_request_params = dict(
            owner='staging',
            query={
                'upload_id': upload['upload_id']
            })

        print('perform repo search on data')
        response = api.post('entries/query', data=json.dumps(query_request_params), auth=auth)
        assert response.status_code == 200, response.text
        response_json = response.json()
        assert response_json['pagination']['total'] == 2
        assert response_json['pagination']['total'] == len(response_json['data'])

        print('performing archive paginated search')
        response = api.post('entries/archive/query', data=json.dumps(dict(
            pagination=dict(page_size=1, page_offset=1),
            **query_request_params)), auth=auth)
        assert response.status_code == 200, response.text
        response_json = response.json()
        assert response_json['pagination']['total'] == 2
        assert len(response_json['data']) == 1

        print('performing archive scrolled search')
        response = api.post('entries/archive/query', data=json.dumps(dict(
            pagination=dict(page_size=1),
            **query_request_params)), auth=auth)
        response_json = response.json()
        response = api.post('entries/archive/query', data=json.dumps(dict(
            pagination=dict(page_size=1, page_after_value=response_json['pagination']['next_page_after_value']),
            **query_request_params)), auth=auth)
        assert response.status_code == 200, response.text
        response_json = response.json()
        assert response_json['pagination']['total'] == 2
        assert len(response_json['data']) == 1

        print('performing download')
        response = api.get(
            'entries/raw',
            params=dict(upload_id=upload['upload_id'], owner='visible'), auth=auth)
        assert response.status_code == 200, response.text

        if not skip_publish:
            print('publish upload')
            api.post(f'uploads/{upload["upload_id"]}/action/publish')

            upload = get_upload(upload)
            assert upload['process_status'] == 'SUCCESS', 'publish must be successful'
            published = True

        print('editing upload')
        response = api.get('users', params=dict(prefix='Markus Scheidgen'))
        assert response.status_code == 200, response.text
        user = response.json()['data'][0]
        dataset = 'test_dataset'
        actions = {
            'comment': {'value': 'Test comment'},
            'references': [{'value': 'http;//test_reference.com'}],
            'entry_coauthors': [{'value': user['user_id']}],
            'datasets': [{'value': dataset}]}

        response = api.post(
            'entries/edit_v0',
            data=json.dumps(dict(actions=actions, **query_request_params)),
            auth=auth)
        assert response.status_code == 200, response.text

        print('list datasets')
        response = api.get('datasets', auth=auth, params=dict(dataset_name=dataset))
        assert response.status_code == 200, response.text
        response_json = response.json()
        assert len(response_json['data']) == 1, response.text
        dataset_id = response_json['data'][0]['dataset_id']

        if not skip_doi and published:
            print('assigning a DOI')
            response = api.post(f'datasets/{dataset_id}/action/doi', auth=auth)
            assert response.status_code == 200, response.text
            has_doi = True

        if not has_doi or auth.user == 'admin':
            print('deleting dataset')
            response = api.delete(f'datasets/{dataset_id}', auth=auth)
            assert response.status_code == 200, response.text

    finally:
        if not published or auth.user == 'admin':
            print('delete the upload again')
            upload = api.delete(f'uploads/{upload["upload_id"]}', auth=auth).json()['data']
            assert get_upload(upload) is None



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click
import logging
import os

from nomad import config, utils


class POPO(dict):
    '''
    A dict subclass that uses attributes as key/value pairs.
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)


@click.group(help=(
    'This is the entry point to nomad\'s command line interface CLI. '
    'It uses a sub-command structure similar to the git command.'))
@click.option('-v', '--verbose', help='sets log level to info', is_flag=True)
@click.option('--debug', help='sets log level to debug', is_flag=True)
@click.option('--log-label', type=str, help='Label applied to logg entries.')
@click.pass_context
def cli(ctx, verbose: bool, debug: bool, log_label: str):
    config.meta.service = os.environ.get('NOMAD_SERVICE', 'cli')
    config.meta.label = log_label

    if debug:
        config.console_log_level = logging.DEBUG
    elif verbose:
        config.console_log_level = logging.INFO
    else:
        config.console_log_level = logging.WARNING
    utils.set_console_log_level(config.console_log_level)


def run_cli():
    try:
        return cli(obj=POPO())  # pylint: disable=E1120,E1123
    except ImportError:
        import sys

        if next(arg for arg in sys.argv if arg == '-v') is not None:
            import traceback
            traceback.print_exc()

        print(
            'You are accessing functionality that requires extra dependencies.\n'
            'Check the NOMAD documentation or install all extra dependencies:\n'
            '  pip install nomad[all]', file=sys.stderr)

        sys.exit(1)
